#include "balmclass.hpp"

namespace balm {

double feat_eigen_limit[2] = {3 * 3, 2 * 2};
double opt_feat_eigen_limit[2] = {4 * 4, 3 * 3};

int OCTO_TREE::voxel_windowsize = 0;

void down_sampling_voxel(pcl::PointCloud<PointType> &pl_feat,
                         double voxel_size) {
  if (voxel_size < 0.01) {
    return;
  }

  unordered_map<VOXEL_LOC, M_POINT, VLocHash, VLocEqual> feat_map;
  uint plsize = pl_feat.size();

  for (uint i = 0; i < plsize; i++) {
    PointType &p_c = pl_feat[i];
    float loc_xyz[3];
    for (int j = 0; j < 3; j++) {
      loc_xyz[j] = p_c.data[j] / voxel_size;
      if (loc_xyz[j] < 0) {
        loc_xyz[j] -= 1.0;
      }
    }

    VOXEL_LOC position((int64_t)loc_xyz[0], (int64_t)loc_xyz[1],
                       (int64_t)loc_xyz[2]);
    auto iter = feat_map.find(position);
    if (iter != feat_map.end()) {
      iter->second.xyz[0] += p_c.x;
      iter->second.xyz[1] += p_c.y;
      iter->second.xyz[2] += p_c.z;
      iter->second.count++;
    } else {
      M_POINT anp;
      anp.xyz[0] = p_c.x;
      anp.xyz[1] = p_c.y;
      anp.xyz[2] = p_c.z;
      anp.count = 1;
      feat_map[position] = anp;
    }
  }

  plsize = feat_map.size();
  pl_feat.clear();
  pl_feat.resize(plsize);

  uint i = 0;
  for (auto iter = feat_map.begin(); iter != feat_map.end(); ++iter) {
    pl_feat[i].x = iter->second.xyz[0] / iter->second.count;
    pl_feat[i].y = iter->second.xyz[1] / iter->second.count;
    pl_feat[i].z = iter->second.xyz[2] / iter->second.count;
    i++;
  }
}

void down_sampling_voxel(PL_VEC &pl_feat, double voxel_size) {
  unordered_map<VOXEL_LOC, M_POINT, VLocHash, VLocEqual> feat_map;
  uint plsize = pl_feat.size();

  for (uint i = 0; i < plsize; i++) {
    Eigen::Vector3d &p_c = pl_feat[i];
    double loc_xyz[3];
    for (int j = 0; j < 3; j++) {
      loc_xyz[j] = p_c[j] / voxel_size;
      if (loc_xyz[j] < 0) {
        loc_xyz[j] -= 1.0;
      }
    }

    VOXEL_LOC position((int64_t)loc_xyz[0], (int64_t)loc_xyz[1],
                       (int64_t)loc_xyz[2]);
    auto iter = feat_map.find(position);
    if (iter != feat_map.end()) {
      iter->second.xyz[0] += p_c[0];
      iter->second.xyz[1] += p_c[1];
      iter->second.xyz[2] += p_c[2];
      iter->second.count++;
    } else {
      M_POINT anp;
      anp.xyz[0] = p_c[0];
      anp.xyz[1] = p_c[1];
      anp.xyz[2] = p_c[2];
      anp.count = 1;
      feat_map[position] = anp;
    }
  }

  plsize = feat_map.size();
  pl_feat.resize(plsize);

  uint i = 0;
  for (auto iter = feat_map.begin(); iter != feat_map.end(); ++iter) {
    pl_feat[i][0] = iter->second.xyz[0] / iter->second.count;
    pl_feat[i][1] = iter->second.xyz[1] / iter->second.count;
    pl_feat[i][2] = iter->second.xyz[2] / iter->second.count;
    i++;
  }
}

void plvec_trans_func(vector<Eigen::Vector3d> &orig,
                      vector<Eigen::Vector3d> &tran, Eigen::Matrix3d R,
                      Eigen::Vector3d t) {
  uint orig_size = orig.size();
  tran.resize(orig_size);

  for (uint i = 0; i < orig_size; i++) {
    tran[i] = R * orig[i] + t;
  }
}

// Convert PointCloud2 to PointType
void rosmsg2ptype(const sensor_msgs::PointCloud2 &pl_msg,
                  pcl::PointCloud<PointType> &plt) {
  pcl::PointCloud<pcl::PointXYZI> pl;
  pcl::fromROSMsg(pl_msg, pl);

  uint asize = pl.size();
  plt.resize(asize);

  for (uint i = 0; i < asize; i++) {
    plt[i].x = pl[i].x;
    plt[i].y = pl[i].y;
    plt[i].z = pl[i].z;
    plt[i].intensity = pl[i].intensity;
  }
}
}