#pragma once

#include "../types/frame.hpp"

namespace xreg {

class RegistrationBase {
 public:
  using Ptr = std::shared_ptr<RegistrationBase>;

  using ConstPtr = std::shared_ptr<const RegistrationBase>;

  EIGEN_MAKE_ALIGNED_OPERATOR_NEW

  RegistrationBase(/* args */);

  virtual ~RegistrationBase() = default;

  void setInputFrames(std::vector<Frame::Ptr> &frames) { frames_ = frames; }

  virtual void initialize();

  virtual void registration();

  void setOverlap(const Eigen::MatrixXd &om);

  static void detectKeyPoints(xreg::PointCloudXYZ::Ptr cloud,
                              xreg::PointCloudXYZ::Ptr &kps,
                              pcl::PointIndices::Ptr &indices);
  static void extractDescriptors(
      pcl::PointCloud<pcl::PointXYZ>::ConstPtr input,
      const pcl::PointCloud<pcl::PointXYZ>::Ptr &keypoints,
      pcl::PointCloud<pcl::FPFHSignature33>::Ptr features);

  // TODO: options
  static void findCorrespondences(
      typename pcl::PointCloud<pcl::FPFHSignature33>::ConstPtr source,
      typename pcl::PointCloud<pcl::FPFHSignature33>::ConstPtr target,
      std::vector<std::pair<int, int>> &corr
      // std::vector<int> &correspondences
  );

 public:
  Eigen::MatrixXd olp_mat_;
  std::vector<Frame::Ptr> frames_;

  std::vector<std::pair<Frame::Ptr, Frame::Ptr>> olp_frames_;
  std::vector<std::vector<std::pair<int, int>>> matches_;

  // std::vector<std::vec>
};

}  // namespace xreg