#pragma once

#include "registration_base.h"

#include "xreg/types/voxel.h"

#include "balm/balmclass.hpp"

namespace xreg {

// using namespace balm;

class RegistrationEFSGD : public RegistrationBase {
 public:
  enum SolveMode {
    GRADIENT_DESCENT_NAIVE = 0,
    GRADIENT_DESCENT_INCR,
    STEEPEST,
    HEAVYBALL,
    MOMENTUM,
    MOMENTUM_SEQ,
    BENGIOS_NAG,
    GRADIENT_DESCENT_BACKTRACKING,
    BFGS
  };

 public:
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW

  RegistrationEFSGD(/* args */);

  virtual ~RegistrationEFSGD() = default;

  void registration() override;

  void initialize() override;

  void scanToMap(Frame::Ptr frame_src);

  // void cut_voxel(
  //     std::unordered_map<VOXEL_LOC, OCTO_TREE *, VLocHash, VLocEqual> &feat_map,
  //     pcl::PointCloud<PointType>::Ptr pl_feat, Eigen::Matrix3d R_p,
  //     Eigen::Vector3d t_p, int feattype, int fnum, int capacity);

  void mapBA(size_t curr_idx);

 public:
  SolveMode solve_mode_ = SolveMode::BENGIOS_NAG;

  std::unique_ptr<VoxelMapOctree> voxel_map_ = nullptr;

  eigen_aligned_std_vector<SE3> Tlb_;

  bool fix_first_ = false;

  eigen_aligned_std_vector<Vector6d> prev_jacobian_, prev_state_;

  bool step_ = false;

  Matrix4d accumulatedQ_;

  double c1_, c2_;
  double alpha_ = 0.2, beta_ = 0.7;
};

}  // namespace xreg