#pragma once

#include "registration_base.h"
#include "xreg/estimation/fast_gicp/ndt_omp.h"

namespace xreg {

struct NDTOptions {
  bool b_scan2map = false;
  std::string method = "ndt";
};

class RegistrationNDT : public RegistrationBase {
 public:
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW

  RegistrationNDT(const std::string &method, NDTOptions options);

  virtual ~RegistrationNDT() = default;

  void initialize() override;

  template <typename PT>
  void scan2scan(pcl::Registration<PT, PT> *reg);

  template <typename PT>
  void scan2map(pcl::Registration<PT, PT> *reg);

  void registration() override;

 public:
  // std::vector<Frame::Ptr> frames;
  /* data */

  eigen_aligned_std_vector<SE3> Tlb_;

  NDTOptions options_;
};

template <>
void RegistrationNDT::scan2scan(pcl::Registration<PointXYZ, PointXYZ> *reg);

template <>
void RegistrationNDT::scan2map(pcl::Registration<PointXYZ, PointXYZ> *reg);

template <>
void RegistrationNDT::scan2scan(
    pcl::Registration<pcl::PointXYZRGBNormal, pcl::PointXYZRGBNormal> *reg);

template <>
void RegistrationNDT::scan2map(
    pcl::Registration<pcl::PointXYZRGBNormal, pcl::PointXYZRGBNormal> *reg);

}  // namespace xreg