#pragma once

#include <ceres/ceres.h>

#include "../types/cell.h"

namespace xreg {

class EigenFactorSQR : public ceres::SizedCostFunction<4, 6> {
public:
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW

public:
  //   EigenFactorSQR(const Matrix4d &LT, const Vector4d &pi);

  EigenFactorSQR(const Matrix4d &lt, const Cell *cell);

  virtual bool Evaluate(double const *const *parameters, double *residuals,
                        double **jacobians) const override;

  // static Eigen::Matrix<double, 2, 3> toSkew(const Vector3d &vec);

  // static Eigen::Matrix<double, 3, 3> skew(const Vector3d &vec);

  // void setTargetPose(const SE3 &tgt);
  //   void setPlaneParam(const Vector4d &pi) { pi_ = pi; }

public:
  const Eigen::Matrix2d sqrt_info_;

  const Eigen::Matrix4d l_;
  //   Vector4d pi_;
  const Cell *cell_ = nullptr;
};

class EigenFactorSQREvalCallBack : public ceres::EvaluationCallback {
public:
  EigenFactorSQREvalCallBack() = default;

  virtual ~EigenFactorSQREvalCallBack() = default;

  virtual void PrepareForEvaluation(bool evaluate_jacobians,
                                    bool new_evaluation_point) override;

  void addPlane(Cell *plane);

  void addFrame(Frame *f, double *factor);

  std::vector<Cell *> cells_;
  std::vector<EigenFactorSQR *> factors_;

  std::vector<Frame *> frames_;
  std::vector<double *> params_;
};
} // namespace xreg