#pragma once

#include <ceres/ceres.h>

#include "../types/cell.h"

namespace xreg {

// residual, rot, trans
class EdgeFactor : public ceres::SizedCostFunction<3, 3, 3> {
public:
  EdgeFactor(Eigen::Vector3d curr_point, Eigen::Vector3d last_point_a,
             Eigen::Vector3d last_point_b);

  virtual ~EdgeFactor() = default;

  virtual bool Evaluate(double const *const *parameters, double *residuals,
                        double **jacobians) const;

  Eigen::Vector3d curr_point_, last_point_a_, last_point_b_;
  double de_norm_;
};

// residual, rot, trans
class SurfPointFactor : public ceres::SizedCostFunction<1, 3, 3> {
public:
  SurfPointFactor(Eigen::Vector3d curr_point, Eigen::Vector3d pt0,
                  Eigen::Vector3d pt1, Eigen::Vector3d pt2);

  virtual ~SurfPointFactor() = default;

  virtual bool Evaluate(double const *const *parameters, double *residuals,
                        double **jacobians) const;

  Eigen::Vector3d pt_, pt0_, pt1_, pt2_;
  Eigen::Vector3d norm_;
};

// residual, rot, trans
// class SurfNormFactor : public ceres::SizedCostFunction<1, 3, 3> {
// public:
//   SurfNormFactor(Eigen::Vector3d curr_point_, Eigen::Vector3d
//   plane_unit_norm_,
//                  double negative_OA_dot_norm_);

//   virtual ~SurfNormFactor() = default;

//   virtual bool Evaluate(double const *const *parameters, double *residuals,
//                         double **jacobians) const;

//   Eigen::Vector3d curr_point_;
//   Eigen::Vector3d plane_unit_norm_;
//   double negative_OA_dot_norm_;
// };
class SurfNormFactorLOAM : public ceres::SizedCostFunction<1, 3, 3> {
public:
  SurfNormFactorLOAM(const Eigen::Vector3d &pt, const Eigen::Vector3d &normal,
                     const double &d);

  virtual ~SurfNormFactorLOAM() = default;

  virtual bool Evaluate(double const *const *parameters, double *residuals,
                        double **jacobians) const;

  Eigen::Vector3d pt_, n_;
  double d_;
};

class SurfNormFactor : public ceres::SizedCostFunction<1, 3, 3> {
public:
  SurfNormFactor(const Eigen::Vector3d &pt, const Eigen::Vector3d &normal,
                 const Eigen::Vector3d &mean);

  virtual ~SurfNormFactor() = default;

  virtual bool Evaluate(double const *const *parameters, double *residuals,
                        double **jacobians) const;

  Eigen::Vector3d pt_, m_, n_;
};

} // namespace xreg
