#pragma once

#include <ceres/ceres.h>

#include "../types/cell.h"

namespace xreg {

class EigenFactorTrans2 : public ceres::SizedCostFunction<1, 3, 3> {
 public:
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW

 public:
  EigenFactorTrans2(const Gaussian3d *info_, const Vector3d &mu,
                    const double s);

  virtual bool Evaluate(double const *const *parameters, double *residuals,
                        double **jacobians) const override;

 public:
  const double scalar_ = 1.0;

  const Gaussian3d *info_map_ = nullptr;
  const Vector3d mu_;
};

class EigenFactorRotAxis : public ceres::SizedCostFunction<1, 3> {
 public:
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW

 public:
  EigenFactorRotAxis(const Gaussian3d *info_map, const Vector3d &axis,
                     const double s);

  virtual bool Evaluate(double const *const *parameters, double *residuals,
                        double **jacobians) const override;

 public:
  const double scalar_ = 1.0;
  const Gaussian3d *info_map_ = nullptr;
  const Vector3d axis_;
};

class SurfRegXXEvalCallBack : public ceres::EvaluationCallback {
 public:
  SurfRegXXEvalCallBack() = default;

  virtual ~SurfRegXXEvalCallBack() = default;

  virtual void PrepareForEvaluation(bool evaluate_jacobians,
                                    bool new_evaluation_point) override;

  void addPlane(Cell *plane);

  void addFrame(Frame *f, double *rot, double *trans);

  std::vector<Cell *> cells_;

  // std::unordered_map<Frame *, double *> x;
  std::vector<Frame *> frames_;
  std::vector<double *> param_rot_, param_trans_;
};

}  // namespace xreg
