#pragma once

#include <ceres/local_parameterization.h>

#include "../common/common.h"

namespace xreg {

struct ParamSO3 : public ceres::LocalParameterization {
  virtual bool Plus(const double *x, const double *delta,
                    double *x_plus_delta) const {
    Eigen::Map<const Vector3d> tangent(x);
    Eigen::Map<const Vector3d> drot(delta);
    Eigen::Map<Vector3d> out(x_plus_delta);

    out = (SO3::exp(tangent) * SO3::exp(drot)).log();

    return true;
  }

  virtual bool ComputeJacobian(const double *x, double *jacobian) const {
    Eigen::Map<Eigen::Matrix<double, 3, 3, Eigen::RowMajor>> j(jacobian);
    j.setIdentity();
    return true;
  }

  virtual int GlobalSize() const { return 3; };
  virtual int LocalSize() const { return 3; };
};

struct ParamSE3 : public ceres::LocalParameterization {
  virtual bool Plus(const double *x, const double *delta,
                    double *x_plus_delta) const {
    Eigen::Map<const Vector6d> tangent(x);
    Eigen::Map<const Vector6d> drot(delta);
    Eigen::Map<Vector6d> out(x_plus_delta);

    out = (SE3::exp(drot) * SE3::exp(tangent)).log();
    // TODO: FUCK?
    // out = (SE3::exp(tangent) * SE3::exp(drot)).log();

    return true;
  }

  virtual bool ComputeJacobian(const double *x, double *jacobian) const {
    Eigen::Map<Eigen::Matrix<double, 6, 6, Eigen::RowMajor>> j(jacobian);
    // SE3::Dxi_exp_x_matrix_at_0
    j.setIdentity();
    return true;
  }

  virtual int GlobalSize() const { return 6; };
  virtual int LocalSize() const { return 6; };
};

} // namespace xreg
