#pragma once

#include "common/common.h"

#include "types/frame.hpp"

#include "visualization/visualizer.h"

#include "estimation/registration_base.h"

namespace xreg {
class System {
public:
  explicit System(ros::NodeHandle nh);

  ~System() = default;

  void loadFrames(const std::string &file_path);

  void visualize();

  void visualizeIdx();

  void spin();

  void shutdown();

  void readTrajectory(const std::string &file_name,
                      eigen_aligned_std_vector<Quaterniond> &rot,
                      eigen_aligned_std_vector<Vector3d> &trans);

  void downsample(PointCloudXYZ::Ptr cloud, PointCloudXYZ::Ptr cloud_filtered,
                  double size);

  void summarize(double time);

public:
  ros::NodeHandle nh_;

  std::vector<Frame::Ptr> frames_;

  RegistrationBase::Ptr registration_ = nullptr;

  Eigen::MatrixXd overlap_matrix_;

  std::unique_ptr<Visualizer> viewer_;
  std::unique_ptr<std::thread> thread_viz_ = nullptr;
};

} // namespace xreg
