#pragma once

#include "../common/common.h"

#include <ros/ros.h>

#include <sensor_msgs/PointCloud2.h>

#include <visualization_msgs/MarkerArray.h>

namespace xreg {

class CloudVisualizer {
public:
  using Ptr = std::unique_ptr<CloudVisualizer>;

  using ConstPtr = std::unique_ptr<const CloudVisualizer>;

  //   void TrajectoryV
  CloudVisualizer() = delete;

  CloudVisualizer(const std::string &name, const std::string &frame_id,
                  ros::NodeHandle &nh);

  ~CloudVisualizer() = default;

  void visualize();

  template <typename PointT>
  void setPointCloud(const pcl::PointCloud<PointT> &cloud);

  std::string name_;
  ros::Publisher pub_;
  std_msgs::Header header_;
  sensor_msgs::PointCloud2::Ptr msg_ = nullptr;
};

class Visualizer {
public:
  using Ptr = std::unique_ptr<Visualizer>;

  using ConstPtr = std::unique_ptr<const Visualizer>;

  explicit Visualizer(ros::NodeHandle &nh);

  virtual ~Visualizer() = default;

  template <typename PointType>
  void addCloud(const std::string &name, const std::string &frame_id,
                const pcl::PointCloud<PointType> &cloud,
                Vector3d offset = Vector3d::Zero());

  void spin();

  void stop() { stop_ = true; }

  void spinOnce();

  void visualizeCloud();

  void broadcastTF();
  // void

  void setTransform(const Quaterniond &rot, const Vector3d &trans,
                    const std::string &name = "camera");

  void setTransform(const SE3 &trans, const std::string &name = "camera");

private:
  ros::NodeHandle nh_;

  std::atomic_bool stop_;
  // ros::Publisher voxel_pub_;

  //   sensor_msgs::PointCloud2::Ptr msg_cloud_;
  std::mutex mutex_cloud_;
  std::unordered_map<std::string, CloudVisualizer::Ptr> cloud_server_;

  // std::unordered_map<std::string, > tf_records_;
  std::mutex mutex_pose_;
  eigen_aligned_std_unordered_map<std::string, SE3> tf_records_;

  //   sensor_msg
};

} // namespace xreg
