# %%
#! /usr/bin/python3

import open3d as o3d

from pathlib import Path
import glob

import numpy as np
import csv

from evo.tools import file_interface

import copy

import matplotlib.cm as cm

# print(cm.hot(0.3))

# gicp
# Transformation is:
# [[ 0.99995574 -0.00927613 -0.00157011 -0.05962241]
#  [ 0.0092812   0.99995163  0.00325402  0.01215244]
#  [ 0.00153984 -0.00326845  0.99999347 -0.01627839]
#  [ 0.          0.          0.          1.        ]]
# 2 922633 0.0011579168855159255
# 2 356774 0.001184870606542293
# <class 'numpy.ndarray'>
# rmse: 0.0644017572660441


base_path = "/home/hyhuang/Data/ethz_registration/"
# dataset = "hauptgebaude"

# data_path = Path(base_path) / dataset
# pcd_path = data_path / "pcd"

datasets = ["apartment", "gazebo_summer", "gazebo_winter",
            "hauptgebaude", "mountain", "stairs", "wood_autumn", "wood_summer"]

log_path = Path(
    "/home/hyhuang/rosWorkSpace/xreg/src/xreg/data/res0624/")
methods = [f for f in log_path.iterdir()]

# est_base_path = "/home/hyhuang/rosWorkSpace/xreg/src/xreg/logs/est_ba_rirhql"
# est_base_path = "/home/hyhuang/rosWorkSpace/xreg/src/xreg/logs/est_ndt_0507_113714"
est_base_path = "/home/hyhuang/rosWorkSpace/xreg/src/xreg/logs/est_gicp_0610_202557"

est_traj_path = Path(est_base_path) / dataset

traj_est = file_interface.read_tum_trajectory_file(est_traj_path)
Test = traj_est.poses_se3


def read_pc_poses(file_name):
    reader = csv.reader(open(file_name, 'r'))
    rows = [row for row in reader]
    rows = rows[1:]

    Tgt = [np.array(list(map(float, row[1:]))).reshape(4, 4) for row in rows]

    return len(Tgt), Tgt
    # pass


num_pcd, Tgt = read_pc_poses(data_path / "icpList.csv")

pcd_data = [o3d.io.read_point_cloud(
    f'{str(pcd_path / str(i))}.pcd') for i in range(num_pcd)]
pcd_data_est = [copy.deepcopy(p) for p in pcd_data]
pcd_gt = [p.transform(T) for (p, T) in zip(pcd_data, Tgt)]
pcd_est = [p.transform(T) for (p, T) in zip(pcd_data_est, Test)]
# pcd_gt = [p.transform(T) for (p, T) in zip(pcd_data, Tgt)]
# print()
# for p in pcd_data:
#     print(np.asarray(p.points).shape)


map_gt = o3d.geometry.PointCloud()
pts_gt = np.vstack([np.asarray(p.points) for p in pcd_gt])
map_gt.points = o3d.utility.Vector3dVector(pts_gt)

map_est = o3d.geometry.PointCloud()
pts_est = np.vstack([np.asarray(p.points) for p in pcd_est])
map_est.points = o3d.utility.Vector3dVector(pts_est)

# map_gt


# %% icp reg
trans_init = np.eye(4)
threshold = 0.05

# reg_p2l = o3d.pipelines.registration.registration_icp(
#     map_est, map_gt, 0.1, trans_init,
#     o3d.pipelines.registration.TransformationEstimationPointToPlane())
# o3d.
reg_p2p = o3d.registration.registration_icp(
    map_est, map_gt, 0.1, trans_init, o3d.registration.TransformationEstimationPointToPoint())
# reg_p2p = o3d.registration.registration_icp(
#     map_est, map_gt, threshold, trans_init,
#     o3d.registration.TransformationEstimationPointToPoint())
# print(reg_p2p)
print("Transformation is:")
print(reg_p2p.transformation)
T = reg_p2p.transformation

# %% compute rmse color map
# T = np.array([9.99953083e-01, 5.98260819e-03,  7.61841956e-03, -2.46448853e-04, -5.96273374e-03, 9.99978767e-01, -2.62878310e-03, -7.06741978e-04, -
#               7.63398478e-03, 2.58323316e-03,  9.99967524e-01,  2.47495373e-02, 0.00000000e+00, 0.00000000e+00,  0.00000000e+00,  1.00000000e+00])

# gicp
import matplotlib.cm as cm

# print(cm.hot(0.3))

# cm.jet()

# T = np.array([0.99995574, -0.00927613, -0.00157011, -0.05962241, 0.0092812, 0.99995163,
#               0.00325402, 0.01215244, 0.00153984, -0.00326845, 0.99999347, -0.01627839, 0., 0., 0., 1])
# T = T.reshape(4, 4)

# T = np.array([9.99950335e-01, -6.15364952e-03, 7.83970549e-03, -3.81453570e-03, 6.14644692e-03, 9.99980666e-01, 9.42496481e-04, -9.61185261e-03, -
#               7.84535371e-03, -8.94263338e-04, 9.99968825e-01, 5.68797019e-03, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 1.00000000e+00])
# T = T.reshape(4, 4)
# print(reg_p2l)
map_est.transform(T)

kdtree = o3d.geometry.KDTreeFlann(map_gt)

# map_est.points
[k, idx, dist] = kdtree.search_knn_vector_3d(map_est.points[0], 2)
print(k, idx[0], dist[0])
print(k, idx[1], dist[1])

dist = map_gt.compute_point_cloud_distance(map_est)
dist = np.asarray(dist)
print(type(dist))

rmse = np.sqrt(np.mean((dist)**2))
print(f'rmse: {rmse}')

max_dist = 0.1
map_est.colors = o3d.utility.Vector3dVector(cm.jet(dist / max_dist)[:, :3])

o3d.visualization.draw_geometries([map_est])

# %%
