# %%
#! /usr/bin/python3

import open3d as o3d

from pathlib import Path
import glob

import numpy as np
import csv

from evo.tools import file_interface

import copy

import matplotlib.cm as cm

import pickle


def read_pc_poses(file_name):
    reader = csv.reader(open(file_name, 'r'))
    rows = [row for row in reader]
    rows = rows[1:]

    Tgt = [np.array(list(map(float, row[1:]))).reshape(4, 4) for row in rows]

    return len(Tgt), Tgt
    # pass

base_path = "/home/hyhuang/Data/ethz_registration/"
datasets = ["apartment", "gazebo_summer", "gazebo_winter",
            "hauptgebaude", "mountain", "stairs", "wood_autumn", "wood_summer"]

# log_path = Path(
#     "/home/hyhuang/rosWorkSpace/xreg/src/xreg/data/result/")
log_path = Path(
    "/home/hyhuang/rosWorkSpace/xreg/src/xreg/data/res0624/")
methods = [f for f in log_path.iterdir()]


for method_path in methods:
    print('###### ', str(method_path).split('/')[-1])
    trans_dict = {}
    for dataset in datasets:
        print('### ', dataset)

        # gt data
        data_path = Path(base_path) / dataset
        pcd_path = data_path / "pcd"

        num_pcd, Tgt = read_pc_poses(data_path / "icpList.csv")

        pcd_data = [o3d.io.read_point_cloud(
            f'{str(pcd_path / str(i))}.pcd') for i in range(num_pcd)]

        # traj_est
        est_traj_path = method_path / dataset
        traj_est = file_interface.read_tum_trajectory_file(est_traj_path)
        Test = traj_est.poses_se3

        pcd_data_est = [copy.deepcopy(p) for p in pcd_data]
        pcd_gt = [p.transform(T) for (p, T) in zip(pcd_data, Tgt)]
        pcd_est = [p.transform(T) for (p, T) in zip(pcd_data_est, Test)]

        map_gt = o3d.geometry.PointCloud()
        pts_gt = np.vstack([np.asarray(p.points) for p in pcd_gt])
        map_gt.points = o3d.utility.Vector3dVector(pts_gt)

        map_est = o3d.geometry.PointCloud()
        pts_est = np.vstack([np.asarray(p.points) for p in pcd_est])
        map_est.points = o3d.utility.Vector3dVector(pts_est)

        # compute transform
        trans_init = np.eye(4)
        threshold = 0.05

        reg_p2p = o3d.registration.registration_icp(
            map_est, map_gt, 0.1, trans_init, o3d.registration.TransformationEstimationPointToPoint())

        print("Transformation is:")
        print(reg_p2p.transformation)
        T = reg_p2p.transformation
        trans_dict[dataset] = T


    with open(method_path / 'trans.pkl', 'wb') as ofs:
        pickle.dump(trans_dict, ofs)
    pass


# %%
