#include "xreg/common/config.h"

#include "xreg/common/common.h"

namespace xreg {

namespace ba {

int idx_ba_init = 1;

bool b_ba_fix_first = false;
int th_ba_min_num_frames_cell = 2;

double voxel_size_surf = 1.0;

bool b_map_check_num = true;
int th_map_min_num = 8;
bool b_map_check_min_lambda = true;
double th_map_min_lambda = 1.0;
bool b_map_check_max_lambda = true;
double th_map_max_lambda = 0.25;
bool b_map_check_p2 = false;
double th_map_p2 = 0.65;

BA_WEIGHT ba_weight_type = BA_WEIGHT::CONSTANT;
BA_LOSS ba_loss_type = BA_LOSS::NONE;
double ba_loss_ratio_trans = 0.1;
double ba_loss_ratio_rot = 0.2;

}  // namespace test_ba

namespace common {

std::string data_path = "/home/hyhuang/Data/ethz_registration/";

std::string dataset = "hauptgebaude";

std::string log_path = "/home/hyhuang/Downloads/";

std::string method = "gicp";

bool b_init = true;
std::string init_method = "teaser";
std::string init_traj_path;

double resolution = 0.4;

double corr_dist = 0.1;

int num_knn = 10;

int num_threads = 1;

int search_method = 0;

bool perform_downsample = false;

float downsample_resolution = 0.05;

int frame_idx = 0;

double overlap_ratio = 0.7;
}  // namespace common

namespace feature {

double resolution_downsample = 0.1;
double resolution_normal = 0.4;
double resolution_keypoints = 0.4;
double resolution_fpfh = 1.6;

bool b_match_all_pairs = false;
}  // namespace feature

namespace tpp {

double noise_bound = 0.1;

double cbar2 = 1;

double gnc_factor = 1.4;

bool b_pgo = false;
}  // namespace tpp

namespace reg {
bool b_scan2map = true;
bool b_fix_first = true;
int num_iteration = 1000;

bool b_downsample = true;
double resolution_downsample = 0.1;
}  // namespace reg

namespace balm {

int window_size = 20;
int marg_size = 1;
double voxel_size = 0.5;
}  // namespace balm

void initParameters(ros::NodeHandle &nh) {
#define GPARAM(x, y)                                  \
  do {                                                \
    if (!nh.getParam(x, y)) {                         \
      LOG(WARNING) << "retrive pararm " #x " error!"; \
    }                                                 \
  } while (0)

  GPARAM("data_path", common::data_path);
  GPARAM("dataset", common::dataset);
  GPARAM("method", common::method);
  GPARAM("log_path", common::log_path);

  GPARAM("b_init", common::b_init);
  GPARAM("init_method", common::init_method);
  GPARAM("init_traj_path", common::init_traj_path);

  GPARAM("search_method", common::search_method);
  GPARAM("resolution", common::resolution);

  GPARAM("corr_dist", common::corr_dist);
  GPARAM("num_knn", common::num_knn);
  GPARAM("num_threads", common::num_threads);

  // TODO: delete
  GPARAM("dresolution", common::downsample_resolution);

  GPARAM("overlap_ratio", common::overlap_ratio);

  GPARAM("idx", common::frame_idx);

  GPARAM("resolution_downsample", feature::resolution_downsample);
  GPARAM("resolution_fpfh", feature::resolution_fpfh);
  GPARAM("resolution_keypoints", feature::resolution_keypoints);
  GPARAM("resolution_normal", feature::resolution_normal);
  GPARAM("match_all_pairs", feature::b_match_all_pairs);

  GPARAM("tpp/noise_bound", tpp::noise_bound);
  GPARAM("tpp/b_pgo", tpp::b_pgo);

  GPARAM("reg/b_scan2map", reg::b_scan2map);
  GPARAM("reg/b_fix_first", reg::b_fix_first);
  GPARAM("reg/num_iteration", reg::num_iteration);
  GPARAM("reg/b_downsample", reg::b_downsample);
  GPARAM("reg/resolution_downsample", reg::resolution_downsample);

  GPARAM("balm/window_size", balm::window_size);
  GPARAM("balm/marg_size", balm::marg_size);
  GPARAM("balm/voxel_size", balm::voxel_size);

  GPARAM("ba/idx_ba_init", ba::idx_ba_init);
  // GPARAM("test_ba/b_ba_fix_first", test_ba::b_ba_fix_first);
  GPARAM("ba/th_ba_min_num_frames_cell", ba::th_ba_min_num_frames_cell);

  GPARAM("ba/voxel_size_surf", ba::voxel_size_surf);
  GPARAM("ba/b_map_check_num", ba::b_map_check_num);
  GPARAM("ba/th_map_min_num", ba::th_map_min_num);
  GPARAM("ba/b_map_check_min_lambda", ba::b_map_check_min_lambda);
  GPARAM("ba/th_map_min_lambda", ba::th_map_min_lambda);
  GPARAM("ba/b_map_check_max_lambda", ba::b_map_check_max_lambda);
  GPARAM("ba/th_map_max_lambda", ba::th_map_max_lambda);
  GPARAM("ba/b_map_check_p2", ba::b_map_check_p2);
  GPARAM("ba/th_map_p2", ba::th_map_p2);

  std::string wt = "constant", lt = "none";
  GPARAM("ba/ba_loss_type", lt);
  GPARAM("ba/ba_weight_type", wt);
  if (lt == "constant") {
    ba::ba_loss_type = BA_LOSS::LCONSTANT;
  } else if (lt == "scalar") {
    ba::ba_loss_type = BA_LOSS::SCALAR_DEPEND;
  }

  if (wt == "local") {
    ba::ba_weight_type = BA_WEIGHT::LOCAL;
  } else if (wt == "global") {
    ba::ba_weight_type = BA_WEIGHT::LOCAL_GLOBAL;
  }

  GPARAM("ba/ba_loss_ratio_trans", ba::ba_loss_ratio_trans);
  GPARAM("ba/ba_loss_ratio_rot", ba::ba_loss_ratio_rot);

#undef GPARAM
}

}  // namespace xreg