#include "xreg/estimation/registration_balm.h"

#include <pcl/filters/voxel_grid.h>

// #include "pcl/registration/icp.h"
#include "xreg/common/config.h"

namespace xreg {
using namespace std;

double voxel_size[2] = {0.5, 0.5};

RegistrationBALM::RegistrationBALM(/* args */) : RegistrationBase() {}

void RegistrationBALM::initialize() {
  INFOLOG("initialize");

  // voxel_map_ =
  //     std::unique_ptr<VoxelMapOctree>(new
  //     VoxelMapOctree(common::resolution));

  voxel_size[0] = balm::voxel_size;
  voxel_size[1] = balm::voxel_size;

  window_size = balm::window_size;
  margi_size = balm::marg_size;

  Tlb_.push_back(SE3());

  for (size_t i = 1; i < frames_.size(); i++) {
    auto &&f = frames_[i];
    auto &&fl = frames_[i - 1];

    // frame
    SE3 Tlb = fl->getTbw() * f->getTwb();
    Tlb_.push_back(Tlb);
  }

  // if (reg::b_downsample) {
  //   INFOLOG("downsample");
  //   pcl::VoxelGrid<pcl::PointXYZ> voxel_filer;
  //   voxel_filer.setLeafSize(reg::resolution_downsample,
  //                           reg::resolution_downsample,
  //                           reg::resolution_downsample);
  //   for (auto &&frame : frames_) {
  //     frame->cloud_down_ = PointCloudXYZ::Ptr(new PointCloudXYZ);

  //     voxel_filer.setInputCloud(frame->cloud_);
  //     voxel_filer.filter(*frame->cloud_down_);
  //   }
  // } else 
  {
    for (auto &&frame : frames_) {
      frame->cloud_down_ = PointCloudXYZ::Ptr(new PointCloudXYZ);
      *frame->cloud_down_ += *frame->cloud_;
    }
  }
}

void RegistrationBALM::cut_voxel(
    unordered_map<VOXEL_LOC, OCTO_TREE *, VLocHash, VLocEqual> &feat_map,
    pcl::PointCloud<PointType>::Ptr pl_feat, Eigen::Matrix3d R_p,
    Eigen::Vector3d t_p, int feattype, int fnum, int capacity) {
  uint plsize = pl_feat->size();
  for (uint i = 0; i < plsize; i++) {
    PointType &p_c = pl_feat->points[i];
    Eigen::Vector3d pvec_orig(p_c.x, p_c.y, p_c.z);
    Eigen::Vector3d pvec_tran = R_p * pvec_orig + t_p;

    float loc_xyz[3];
    for (int j = 0; j < 3; j++) {
      loc_xyz[j] = pvec_tran[j] / voxel_size[feattype];
      if (loc_xyz[j] < 0) {
        loc_xyz[j] -= 1.0;
      }
    }

    VOXEL_LOC position((int64_t)loc_xyz[0], (int64_t)loc_xyz[1],
                       (int64_t)loc_xyz[2]);
    auto iter = feat_map.find(position);
    if (iter != feat_map.end()) {
      iter->second->plvec_orig[fnum]->push_back(pvec_orig);
      iter->second->plvec_tran[fnum]->push_back(pvec_tran);
      iter->second->is2opt = true;
    } else {
      OCTO_TREE *ot = new OCTO_TREE(feattype, capacity);
      ot->plvec_orig[fnum]->push_back(pvec_orig);
      ot->plvec_tran[fnum]->push_back(pvec_tran);

      ot->voxel_center[0] = (0.5 + position.x) * voxel_size[feattype];
      ot->voxel_center[1] = (0.5 + position.y) * voxel_size[feattype];
      ot->voxel_center[2] = (0.5 + position.z) * voxel_size[feattype];
      ot->quater_length = voxel_size[feattype] / 4.0;
      feat_map[position] = ot;
    }
  }
}

void RegistrationBALM::registration() {
  int jump_flag = skip_num;
  printf("%d\n", skip_num);
  LM_SLWD_VOXEL opt_lsv(window_size, filter_num, thread_num);

  // pcl::PointCloud<PointType>::Ptr pl_corn(new pcl::PointCloud<PointType>);
  pcl::PointCloud<PointType>::Ptr pl_surf(new pcl::PointCloud<PointType>);
  vector<pcl::PointCloud<PointType>::Ptr> pl_full_buf;

  vector<Eigen::Quaterniond> q_poses;
  vector<Eigen::Vector3d> t_poses;
  // Eigen::Quaterniond q_odom, q_gather_pose(1, 0, 0, 0), q_last(1, 0, 0, 0);
  // Eigen::Vector3d t_odom, t_gather_pose(0, 0, 0), t_last(0, 0, 0);

  // plcount ==> current index
  int plcount = 0, window_base = 0;

  // unordered_map<VOXEL_LOC, OCTO_TREE *, VLocHash, VLocEqual> surf_map,
  // corn_map;
  unordered_map<VOXEL_LOC, OCTO_TREE *, VLocHash, VLocEqual> surf_map;
  Eigen::Matrix4d trans(Eigen::Matrix4d::Identity());
  // geometry_msgs::PoseArray parray;
  // parray.header.frame_id = "camera_init";

  // while (n.ok())

  auto x2pttype = [](PointCloudXYZ::Ptr pl, pcl::PointCloud<PointType> &plt) {
    uint asize = pl->size();
    plt.resize(asize);

    for (uint i = 0; i < asize; i++) {
      plt[i].x = pl->points[i].x;
      plt[i].y = pl->points[i].y;
      plt[i].z = pl->points[i].z;
      // plt[i].intensity = pl[i].intensity;
    }
  };

  for (size_t i = 0; i < frames_.size(); i++) {
    // STEP. create compatible point cloud type for BALM
    auto &&frame = frames_[i];

    // (frame->cloud_down_;
    x2pttype(frame->cloud_down_, *pl_surf);

    // corn_buf.pop();
    // surf_buf.pop();
    // full_buf.pop();

    // q_odom.w() = odom_buf.front()->pose.pose.orientation.w;
    // q_odom.x() = odom_buf.front()->pose.pose.orientation.x;
    // q_odom.y() = odom_buf.front()->pose.pose.orientation.y;
    // q_odom.z() = odom_buf.front()->pose.pose.orientation.z;
    // t_odom.x() = odom_buf.front()->pose.pose.position.x;
    // t_odom.y() = odom_buf.front()->pose.pose.position.y;
    // t_odom.z() = odom_buf.front()->pose.pose.position.z;
    // odom_buf.pop();
    // mBuf.unlock();

    // Eigen::Vector3d delta_t(q_last.matrix().transpose() * (t_odom -
    // t_last)); Eigen::Quaterniond delta_q(q_last.matrix().transpose() *
    // q_odom.matrix()); q_last = q_odom; t_last = t_odom;

    // t_gather_pose = t_gather_pose + q_gather_pose * delta_t;
    // q_gather_pose = q_gather_pose * delta_q;
    // if (jump_flag < skip_num) {
    //   jump_flag++;
    //   continue;
    // }
    // jump_flag = 0;

    // initial pose
    if (plcount == 0) {
      q_poses.push_back(Quaterniond::Identity());
      t_poses.push_back(Vector3d::Zero());
    } else {
      SE3 Tlb = Tlb_[i];
      t_poses.push_back(t_poses[plcount - 1] +
                        q_poses[plcount - 1] * Tlb.translation());
      q_poses.push_back(q_poses[plcount - 1] * Tlb.unit_quaternion());
      // SE3 Tgt = frame->Tgt_;
      // t_poses.push_back(Tgt.translation());
      // q_poses.push_back(Tgt.unit_quaternion());
    }

    // parray.header.stamp = ct;
    // geometry_msgs::Pose apose;
    // apose.orientation.w = q_poses[plcount].w();
    // apose.orientation.x = q_poses[plcount].x();
    // apose.orientation.y = q_poses[plcount].y();
    // apose.orientation.z = q_poses[plcount].z();
    // apose.position.x = t_poses[plcount].x();
    // apose.position.y = t_poses[plcount].y();
    // apose.position.z = t_poses[plcount].z();
    // parray.poses.push_back(apose);
    // pub_pose.publish(parray);

    // pl_full_buf.push_back(pl_full);
    plcount++;

    OCTO_TREE::voxel_windowsize = plcount - window_base;
    // q_gather_pose.setIdentity();
    // t_gather_pose.setZero();

    // down_sampling_voxel(*pl_corn, corn_filter_length);
    if (reg::b_downsample) {
      down_sampling_voxel(*pl_surf, reg::resolution_downsample);
    }

    int frame_head = plcount - 1 - window_base;
    cut_voxel(surf_map, pl_surf, q_poses[plcount - 1].matrix(),
              t_poses[plcount - 1], 0, frame_head, window_size);
    // cut_voxel(corn_map, pl_corn, q_poses[plcount - 1].matrix(),
    //           t_poses[plcount - 1], 1, frame_head, window_size);

    for (auto iter = surf_map.begin(); iter != surf_map.end(); ++iter) {
      if (iter->second->is2opt) {
        iter->second->root_centors.clear();
        iter->second->recut(0, frame_head, iter->second->root_centors);
      }
    }

    // for (auto iter = corn_map.begin(); iter != corn_map.end(); ++iter) {
    //   if (iter->second->is2opt) {
    //     iter->second->root_centors.clear();
    //     iter->second->recut(0, frame_head, iter->second->root_centors);
    //   }
    // }

    if (plcount >= window_base + window_size) {
      for (int i = 0; i < window_size; i++) {
        opt_lsv.so3_poses[i].setQuaternion(q_poses[window_base + i]);
        opt_lsv.t_poses[i] = t_poses[window_base + i];
      }

      if (window_base != 0) {
        for (auto iter = surf_map.begin(); iter != surf_map.end(); ++iter) {
          if (iter->second->is2opt) {
            iter->second->traversal_opt(opt_lsv);
          }
        }

        // for (auto iter = corn_map.begin(); iter != corn_map.end(); ++iter)
        // {
        //   if (iter->second->is2opt) {
        //     iter->second->traversal_opt(opt_lsv);
        //   }
        // }

        INFOLOG("ba iter");
        opt_lsv.damping_iter();
      }

      // pcl::PointCloud<PointType> pl_send;

      // for (int i = 0; i < margi_size; i += pub_skip) {
      //   trans.block<3, 3>(0, 0) = opt_lsv.so3_poses[i].matrix();
      //   trans.block<3, 1>(0, 3) = opt_lsv.t_poses[i];

      //   pcl::PointCloud<PointType> pcloud;
      //   pcl::transformPointCloud(*pl_full_buf[window_base + i], pcloud,
      //   trans); pl_send += pcloud;
      // }

      // pub_func(pl_send, pub_full, ct);

      // for (int i = 0; i < margi_size; i++) {
      //   pl_full_buf[window_base + i] = nullptr;
      // }

      for (int i = 0; i < window_size; i++) {
        q_poses[window_base + i] = opt_lsv.so3_poses[i].unit_quaternion();
        t_poses[window_base + i] = opt_lsv.t_poses[i];
      }

      // for (int i = window_base; i < window_base + window_size; i++) {
      //   parray.poses[i].orientation.w = q_poses[i].w();
      //   parray.poses[i].orientation.x = q_poses[i].x();
      //   parray.poses[i].orientation.y = q_poses[i].y();
      //   parray.poses[i].orientation.z = q_poses[i].z();
      //   parray.poses[i].position.x = t_poses[i].x();
      //   parray.poses[i].position.y = t_poses[i].y();
      //   parray.poses[i].position.z = t_poses[i].z();
      // }
      // pub_pose.publish(parray);

      for (auto iter = surf_map.begin(); iter != surf_map.end(); ++iter) {
        if (iter->second->is2opt) {
          iter->second->root_centors.clear();
          iter->second->marginalize(0, margi_size, q_poses, t_poses,
                                    window_base, iter->second->root_centors);
        }
      }

      // for (auto iter = corn_map.begin(); iter != corn_map.end(); ++iter) {
      //   if (iter->second->is2opt) {
      //     iter->second->root_centors.clear();
      //     iter->second->marginalize(0, margi_size, q_poses, t_poses,
      //                               window_base,
      //                               iter->second->root_centors);
      //   }
      // }

      window_base += margi_size;
      opt_lsv.free_voxel();
    }
  }
  // INFOLOG("?");

  // cout << q_poses.size() << endl;
  CHECK_EQ(frames_.size(), q_poses.size());
  for (size_t i = 0; i < frames_.size(); i++) {
    // aut
    frames_[i]->setTwb(SE3(q_poses[i], t_poses[i]));
  }
}

}  // namespace xreg