
#include "xreg/estimation/registration_base.h"

#include "pcl/common/common.h"
#include "pcl/common/transforms.h"
#include "pcl/filters/voxel_grid.h"

#include <pcl/filters/radius_outlier_removal.h>

#include "xreg/common/config.h"

#include <omp.h>

namespace xreg {

using namespace std;

// TODO: hardcode
// constexpr double downsample_resolution = 0.1;
// constexpr double overlap_ratio = 0.5;

RegistrationBase::RegistrationBase(/* args */) {}

void RegistrationBase::initialize() {
  // STEP. set initial value
  for (size_t i = 0; i < frames_.size(); i++) {
    frames_[i]->setTwb(frames_[i]->Tgt_);
  }

  // STEP. downsample
  {
    pcl::VoxelGrid<pcl::PointXYZ> voxel_filer;
    voxel_filer.setLeafSize(feature::resolution_downsample,
                            feature::resolution_downsample,
                            feature::resolution_downsample);
    for (auto &&frame : frames_) {
      frame->cloud_down_ = PointCloudXYZ::Ptr(new PointCloudXYZ);

      voxel_filer.setInputCloud(frame->cloud_);
      voxel_filer.filter(*frame->cloud_down_);
    }
  }

  // STEP. compute 3d bbox
  {
    for (auto &&frame : frames_) {
      pcl::PointXYZ pt_min, pt_max;

      // pcl::PointCloud
      PointCloudXYZ::Ptr cloud_trans(new PointCloudXYZ);
      pcl::transformPointCloud(*frame->cloud_, *cloud_trans,
                               frame->Tgt_.matrix());

      pcl::getMinMax3D(*cloud_trans, pt_min, pt_max);
      frame->setBBox(pt_min, pt_max);
    }

    if (feature::b_match_all_pairs) {
      CHECK_EQ(olp_mat_.cols(), frames_.size());
      for (size_t i = 0; i < frames_.size(); i++) {
        auto fi = frames_[i];

        // auto bboxi = fi->getBBox();
        for (size_t j = i + 1; j < frames_.size(); j++) {
          auto fj = frames_[j];

          if (olp_mat_(i, j) > common::overlap_ratio ||
              olp_mat_(j, i) > common::overlap_ratio) {
            olp_frames_.push_back(make_pair(fi, fj));
          }
          // auto bboxj = fj->getBBox();
          // fj->

          // bounds_t::get_intersection_bbx(bboxi, bboxj, );
          // auto ratio = bboxi.get_intersection_ratio(bboxj);
          // INFOLOG("pair: {}, {}, ratio: {}", i, j, ratio);
          // double ratio =
          //     (fi->Tgt_.translation() - fj->Tgt_.translation()).norm();

          // // intersected
          // if (ratio < common::overlap_ratio) {
          //   olp_frames_.push_back(make_pair(fi, fj));
          // }
        }
      }
    } else {
      for (size_t i = 0; i < frames_.size() - 1; i++) {
        auto fi = frames_[i];
        auto fj = frames_[i + 1];
        olp_frames_.push_back(make_pair(fi, fj));
      }
    }
    INFOLOG("#overlapped pairs: {}", olp_frames_.size());
  }

  // STEP. extract keypoints && features
  {
    INFOLOG("extract keypoints and features");
    for (auto &&frame : frames_) {
      frame->kpts_ = PointCloudXYZ::Ptr(new PointCloudXYZ);
      frame->features_ = CloudFPFH::Ptr(new CloudFPFH);

      pcl::PointIndicesPtr inds(new pcl::PointIndices);
      detectKeyPoints(frame->cloud_down_, frame->kpts_, inds);

      extractDescriptors(frame->cloud_down_, frame->kpts_, frame->features_);

      INFOLOG("frame {}, #kps: {}", 0, frame->kpts_->size());
    }

    INFOLOG("done kps extraction.");
  }

  // STEP. find matches
  {
    matches_.resize(olp_frames_.size());
// for (auto &&pair : olp_frames_)
#pragma omp parallel for num_threads(7)
    for (size_t i = 0; i < olp_frames_.size(); i++) {
      auto &&pair = olp_frames_[i];
      auto fi = pair.first;
      auto fj = pair.second;

      std::vector<std::pair<int, int>> corr;
      findCorrespondences(fj->features_, fi->features_, corr);

// matches_.push_back(corr);
#pragma omp critical
      matches_[i] = corr;
    }
  }
}

void RegistrationBase::setOverlap(const Eigen::MatrixXd &om) { olp_mat_ = om; }

void RegistrationBase::registration() {
  throw std::runtime_error("not implemented");
}

}  // namespace xreg