#include "xreg/feature/feature_extraction.h"

#include <pcl/filters/filter.h>
#include <pcl/filters/voxel_grid.h>

namespace xreg {
// TODO: is neighborhood info important?

FeatureExtractor::FeatureExtractor(const FeatureExtractor::Param &param)
    : param_(param) {

  feats_flat_ = PointCloudXYZI::Ptr(new PointCloudXYZI);
  feats_less_flat_ = PointCloudXYZI::Ptr(new PointCloudXYZI);
  feats_less_flat_raw_ = PointCloudXYZI::Ptr(new PointCloudXYZI);
  feats_sharp_ = PointCloudXYZI::Ptr(new PointCloudXYZI);
  feats_less_sharp_ = PointCloudXYZI::Ptr(new PointCloudXYZI);
}

float cloudCurvature[400000];
int cloudSortInd[400000];
int cloudNeighborPicked[400000];
int cloudLabel[400000];

bool comp(int i, int j) { return (cloudCurvature[i] < cloudCurvature[j]); }

void FeatureExtractor::extractALOAM(const PointCloudXYZI::Ptr &pc_in) {
  feats_flat_ = PointCloudXYZI::Ptr(new PointCloudXYZI);
  feats_less_flat_ = PointCloudXYZI::Ptr(new PointCloudXYZI);
  feats_less_flat_raw_ = PointCloudXYZI::Ptr(new PointCloudXYZI);
  feats_sharp_ = PointCloudXYZI::Ptr(new PointCloudXYZI);
  feats_less_sharp_ = PointCloudXYZI::Ptr(new PointCloudXYZI);
  // feats_flat_->clear();
  // feats_less_flat_->clear();
  // feats_less_flat_raw_->clear();
  // feats_sharp_->clear();
  // feats_less_sharp_->clear();

  std::vector<int> scanStartInd(param_.num_lines, 0);
  std::vector<int> scanEndInd(param_.num_lines, 0);

  std::vector<int> indices;
  PointCloudXYZI &laserCloudIn = *pc_in;
  pcl::removeNaNFromPointCloud(laserCloudIn, laserCloudIn, indices);
  // removeClosedPointCloud(laserCloudIn, laserCloudIn, MINIMUM_RANGE);

  int cloudSize = laserCloudIn.points.size();
  float startOri = -atan2(laserCloudIn.points[0].y, laserCloudIn.points[0].x);
  float endOri = -atan2(laserCloudIn.points[cloudSize - 1].y,
                        laserCloudIn.points[cloudSize - 1].x) +
                 2 * M_PI;

  if (endOri - startOri > 3 * M_PI) {
    endOri -= 2 * M_PI;
  } else if (endOri - startOri < M_PI) {
    endOri += 2 * M_PI;
  }
  // printf("end Ori %f\n", endOri);

  bool halfPassed = false;
  int count = cloudSize;
  PointXYZI point;
  std::vector<pcl::PointCloud<PointXYZI>> laserCloudScans(param_.num_lines);
  for (int i = 0; i < cloudSize; i++) {
    point.x = laserCloudIn.points[i].x;
    point.y = laserCloudIn.points[i].y;
    point.z = laserCloudIn.points[i].z;

    float angle = atan(point.z / sqrt(point.x * point.x + point.y * point.y)) *
                  180 / M_PI;
    int scanID = 0;

    if (param_.num_lines == 16) {
      scanID = int((angle + 15) / 2 + 0.5);
      if (scanID > (param_.num_lines - 1) || scanID < 0) {
        count--;
        continue;
      }
    } else if (param_.num_lines == 32) {
      scanID = int((angle + 92.0 / 3.0) * 3.0 / 4.0);
      if (scanID > (param_.num_lines - 1) || scanID < 0) {
        count--;
        continue;
      }
    } else if (param_.num_lines == 64) {
      if (angle >= -8.83)
        scanID = int((2 - angle) * 3.0 + 0.5);
      else
        scanID = param_.num_lines / 2 + int((-8.83 - angle) * 2.0 + 0.5);

      // use [0 50]  > 50 remove outlies
      if (angle > 2 || angle < -24.33 || scanID > 50 || scanID < 0) {
        count--;
        continue;
      }
    } else {
      printf("wrong scan number\n");
      ROS_BREAK();
    }
    // printf("angle %f scanID %d \n", angle, scanID);

    float ori = -atan2(point.y, point.x);
    if (!halfPassed) {
      if (ori < startOri - M_PI / 2) {
        ori += 2 * M_PI;
      } else if (ori > startOri + M_PI * 3 / 2) {
        ori -= 2 * M_PI;
      }

      if (ori - startOri > M_PI) {
        halfPassed = true;
      }
    } else {
      ori += 2 * M_PI;
      if (ori < endOri - M_PI * 3 / 2) {
        ori += 2 * M_PI;
      } else if (ori > endOri + M_PI / 2) {
        ori -= 2 * M_PI;
      }
    }

    float relTime = (ori - startOri) / (endOri - startOri);
    point.intensity = scanID + param_.scan_period * relTime;
    laserCloudScans[scanID].push_back(point);
  }

  cloudSize = count;
  printf("points size %d \n", cloudSize);

  pcl::PointCloud<PointXYZI>::Ptr laserCloud(new pcl::PointCloud<PointXYZI>());
  for (int i = 0; i < param_.num_lines; i++) {
    scanStartInd[i] = laserCloud->size() + 5;
    *laserCloud += laserCloudScans[i];
    scanEndInd[i] = laserCloud->size() - 6;
  }

  // printf("prepare time %f \n", t_prepare.toc());

  for (int i = 5; i < cloudSize - 5; i++) {
    float diffX = laserCloud->points[i - 5].x + laserCloud->points[i - 4].x +
                  laserCloud->points[i - 3].x + laserCloud->points[i - 2].x +
                  laserCloud->points[i - 1].x - 10 * laserCloud->points[i].x +
                  laserCloud->points[i + 1].x + laserCloud->points[i + 2].x +
                  laserCloud->points[i + 3].x + laserCloud->points[i + 4].x +
                  laserCloud->points[i + 5].x;
    float diffY = laserCloud->points[i - 5].y + laserCloud->points[i - 4].y +
                  laserCloud->points[i - 3].y + laserCloud->points[i - 2].y +
                  laserCloud->points[i - 1].y - 10 * laserCloud->points[i].y +
                  laserCloud->points[i + 1].y + laserCloud->points[i + 2].y +
                  laserCloud->points[i + 3].y + laserCloud->points[i + 4].y +
                  laserCloud->points[i + 5].y;
    float diffZ = laserCloud->points[i - 5].z + laserCloud->points[i - 4].z +
                  laserCloud->points[i - 3].z + laserCloud->points[i - 2].z +
                  laserCloud->points[i - 1].z - 10 * laserCloud->points[i].z +
                  laserCloud->points[i + 1].z + laserCloud->points[i + 2].z +
                  laserCloud->points[i + 3].z + laserCloud->points[i + 4].z +
                  laserCloud->points[i + 5].z;

    cloudCurvature[i] = diffX * diffX + diffY * diffY + diffZ * diffZ;
    cloudSortInd[i] = i;
    cloudNeighborPicked[i] = 0;
    cloudLabel[i] = 0;
  }

  // pcl::PointCloud<PointXYZI> cornerPointsSharp;
  // pcl::PointCloud<PointXYZI> cornerPointsLessSharp;
  // pcl::PointCloud<PointXYZI> surfPointsFlat;
  // pcl::PointCloud<PointXYZI> surfPointsLessFlat;

  float t_q_sort = 0;

  for (int i = 0; i < param_.num_lines; i++) {
    if (scanEndInd[i] - scanStartInd[i] < 6)
      continue;
    pcl::PointCloud<PointXYZI>::Ptr surfPointsLessFlatScan(
        new pcl::PointCloud<PointXYZI>);
    for (int j = 0; j < 6; j++) {
      int sp = scanStartInd[i] + (scanEndInd[i] - scanStartInd[i]) * j / 6;
      int ep =
          scanStartInd[i] + (scanEndInd[i] - scanStartInd[i]) * (j + 1) / 6 - 1;

      std::sort(cloudSortInd + sp, cloudSortInd + ep + 1, comp);

      int largestPickedNum = 0;
      for (int k = ep; k >= sp; k--) {
        int ind = cloudSortInd[k];

        if (cloudNeighborPicked[ind] == 0 && cloudCurvature[ind] > 0.1) {

          largestPickedNum++;
          if (largestPickedNum <= 2) {
            cloudLabel[ind] = 2;
            feats_sharp_->push_back(laserCloud->points[ind]);
            feats_less_sharp_->push_back(laserCloud->points[ind]);
          } else if (largestPickedNum <= 20) {
            cloudLabel[ind] = 1;
            // cornerPointsLessSharp.push_back(laserCloud->points[ind]);
            feats_less_sharp_->push_back(laserCloud->points[ind]);
          } else {
            break;
          }

          cloudNeighborPicked[ind] = 1;

          for (int l = 1; l <= 5; l++) {
            float diffX = laserCloud->points[ind + l].x -
                          laserCloud->points[ind + l - 1].x;
            float diffY = laserCloud->points[ind + l].y -
                          laserCloud->points[ind + l - 1].y;
            float diffZ = laserCloud->points[ind + l].z -
                          laserCloud->points[ind + l - 1].z;
            if (diffX * diffX + diffY * diffY + diffZ * diffZ > 0.05) {
              break;
            }

            cloudNeighborPicked[ind + l] = 1;
          }
          for (int l = -1; l >= -5; l--) {
            float diffX = laserCloud->points[ind + l].x -
                          laserCloud->points[ind + l + 1].x;
            float diffY = laserCloud->points[ind + l].y -
                          laserCloud->points[ind + l + 1].y;
            float diffZ = laserCloud->points[ind + l].z -
                          laserCloud->points[ind + l + 1].z;
            if (diffX * diffX + diffY * diffY + diffZ * diffZ > 0.05) {
              break;
            }

            cloudNeighborPicked[ind + l] = 1;
          }
        }
      }

      int smallestPickedNum = 0;
      for (int k = sp; k <= ep; k++) {
        int ind = cloudSortInd[k];

        if (cloudNeighborPicked[ind] == 0 && cloudCurvature[ind] < 0.1) {

          cloudLabel[ind] = -1;
          // surfPointsFlat.push_back(laserCloud->points[ind]);
          feats_flat_->push_back(laserCloud->points[ind]);

          smallestPickedNum++;
          if (smallestPickedNum >= 4) {
            break;
          }

          cloudNeighborPicked[ind] = 1;
          for (int l = 1; l <= 5; l++) {
            float diffX = laserCloud->points[ind + l].x -
                          laserCloud->points[ind + l - 1].x;
            float diffY = laserCloud->points[ind + l].y -
                          laserCloud->points[ind + l - 1].y;
            float diffZ = laserCloud->points[ind + l].z -
                          laserCloud->points[ind + l - 1].z;
            if (diffX * diffX + diffY * diffY + diffZ * diffZ > 0.05) {
              break;
            }

            cloudNeighborPicked[ind + l] = 1;
          }
          for (int l = -1; l >= -5; l--) {
            float diffX = laserCloud->points[ind + l].x -
                          laserCloud->points[ind + l + 1].x;
            float diffY = laserCloud->points[ind + l].y -
                          laserCloud->points[ind + l + 1].y;
            float diffZ = laserCloud->points[ind + l].z -
                          laserCloud->points[ind + l + 1].z;
            if (diffX * diffX + diffY * diffY + diffZ * diffZ > 0.05) {
              break;
            }

            cloudNeighborPicked[ind + l] = 1;
          }
        }
      }

      for (int k = sp; k <= ep; k++) {
        if (cloudLabel[k] <= 0) {
          surfPointsLessFlatScan->push_back(laserCloud->points[k]);
        }
      }
    }

    pcl::PointCloud<PointXYZI> surfPointsLessFlatScanDS;
    pcl::VoxelGrid<PointXYZI> downSizeFilter;
    downSizeFilter.setInputCloud(surfPointsLessFlatScan);
    downSizeFilter.setLeafSize(0.2, 0.2, 0.2);
    downSizeFilter.filter(surfPointsLessFlatScanDS);

    *feats_less_flat_raw_ += *surfPointsLessFlatScan;
    *feats_less_flat_ += surfPointsLessFlatScanDS;
  }
  // printf("sort q time %f \n", t_q_sort);
  // printf("seperate points time %f \n", t_pts.toc());
}

void FeatureExtractor::extract(const PointCloudXYZI::Ptr &pc_in) {
  // pc
  // std::vector<int> indices;
  // pcl::removeNaNFromPointCloud(*pc_in, indices);
  // feats_flat_->clear();
  // feats_less_flat_->clear();
  // feats_sharp_->clear();
  // feats_less_sharp_->clear();
  // feats_flat_ = PointCloudXYZI::Ptr(new PointCloudXYZI);
  // feats_less_flat_ = PointCloudXYZI::Ptr(new PointCloudXYZI);
  // feats_less_flat_raw_ = PointCloudXYZI::Ptr(new PointCloudXYZI);
  // feats_sharp_ = PointCloudXYZI::Ptr(new PointCloudXYZI);
  // feats_less_sharp_ = PointCloudXYZI::Ptr(new PointCloudXYZI);
  feats_flat_->clear();
  feats_less_flat_->clear();
  feats_less_flat_raw_->clear();
  feats_sharp_->clear();
  feats_less_sharp_->clear();

  // int param_.num_lines = param_.num_lines;
  std::vector<pcl::PointCloud<pcl::PointXYZI>::Ptr> laserCloudScans;
  for (int i = 0; i < param_.num_lines; i++) {
    laserCloudScans.push_back(pcl::PointCloud<pcl::PointXYZI>::Ptr(
        new pcl::PointCloud<pcl::PointXYZI>()));
  }

  int num_pts = pc_in->points.size();
  float startOri = -atan2(pc_in->points[0].y, pc_in->points[0].x);
  float endOri =
      -atan2(pc_in->points[num_pts - 1].y, pc_in->points[num_pts - 1].x) +
      2 * M_PI;

  if (endOri - startOri > 3 * M_PI) {
    endOri -= 2 * M_PI;
  } else if (endOri - startOri < M_PI) {
    endOri += 2 * M_PI;
  }

  bool halfPassed = false;
  PointXYZI pt_tmp;
  for (int i = 0; i < num_pts; i++) {
    int scanID = 0;
    pt_tmp = pc_in->points[i];
    double distance = sqrt(pt_tmp.x * pt_tmp.x + pt_tmp.y * pt_tmp.y);
    if (distance < param_.min_distance || distance > param_.max_distance)
      continue;

    double angle = atan(pc_in->points[i].z / distance) * 180 / M_PI;

    if (param_.num_lines == 16) {
      scanID = int((angle + 15) / 2 + 0.5);
      if (scanID > (param_.num_lines - 1) || scanID < 0) {
        continue;
      }
    } else if (param_.num_lines == 32) {
      scanID = int((angle + 92.0 / 3.0) * 3.0 / 4.0);
      if (scanID > (param_.num_lines - 1) || scanID < 0) {
        continue;
      }
    } else if (param_.num_lines == 64) {
      if (angle >= -8.83)
        scanID = int((2 - angle) * 3.0 + 0.5);
      else
        scanID = param_.num_lines / 2 + int((-8.83 - angle) * 2.0 + 0.5);

      if (angle > 2 || angle < -24.33 || scanID > 63 || scanID < 0) {
        continue;
      }
    } else {
      //   printf("wrong scan number\n");
      ERRLOG("wrong scan number");
    }

    // ADD(hhy) for relative time calculation
    float ori = -atan2(pt_tmp.y, pt_tmp.x);
    if (!halfPassed) {
      if (ori < startOri - M_PI / 2) {
        ori += 2 * M_PI;
      } else if (ori > startOri + M_PI * 3 / 2) {
        ori -= 2 * M_PI;
      }

      if (ori - startOri > M_PI) {
        halfPassed = true;
      }
    } else {
      ori += 2 * M_PI;
      if (ori < endOri - M_PI * 3 / 2) {
        ori += 2 * M_PI;
      } else if (ori > endOri + M_PI / 2) {
        ori -= 2 * M_PI;
      }
    }

    float relTime = (ori - startOri) / (endOri - startOri);
    pt_tmp.intensity = scanID + param_.scan_period * relTime;

    laserCloudScans[scanID]->push_back(pt_tmp);
  }

  for (int i = 0; i < param_.num_lines; i++) {
    if (laserCloudScans[i]->points.size() < 131) {
      // TODO: why?
      WARNLOG("few points in a scan");
      continue;
    }

    std::vector<IdValue> cloudCurvature;
    int total_points = laserCloudScans[i]->points.size() - 10;
    for (int j = 5; j < (int)laserCloudScans[i]->points.size() - 5; j++) {
      double diffX = laserCloudScans[i]->points[j - 5].x +
                     laserCloudScans[i]->points[j - 4].x +
                     laserCloudScans[i]->points[j - 3].x +
                     laserCloudScans[i]->points[j - 2].x +
                     laserCloudScans[i]->points[j - 1].x -
                     10 * laserCloudScans[i]->points[j].x +
                     laserCloudScans[i]->points[j + 1].x +
                     laserCloudScans[i]->points[j + 2].x +
                     laserCloudScans[i]->points[j + 3].x +
                     laserCloudScans[i]->points[j + 4].x +
                     laserCloudScans[i]->points[j + 5].x;
      double diffY = laserCloudScans[i]->points[j - 5].y +
                     laserCloudScans[i]->points[j - 4].y +
                     laserCloudScans[i]->points[j - 3].y +
                     laserCloudScans[i]->points[j - 2].y +
                     laserCloudScans[i]->points[j - 1].y -
                     10 * laserCloudScans[i]->points[j].y +
                     laserCloudScans[i]->points[j + 1].y +
                     laserCloudScans[i]->points[j + 2].y +
                     laserCloudScans[i]->points[j + 3].y +
                     laserCloudScans[i]->points[j + 4].y +
                     laserCloudScans[i]->points[j + 5].y;
      double diffZ = laserCloudScans[i]->points[j - 5].z +
                     laserCloudScans[i]->points[j - 4].z +
                     laserCloudScans[i]->points[j - 3].z +
                     laserCloudScans[i]->points[j - 2].z +
                     laserCloudScans[i]->points[j - 1].z -
                     10 * laserCloudScans[i]->points[j].z +
                     laserCloudScans[i]->points[j + 1].z +
                     laserCloudScans[i]->points[j + 2].z +
                     laserCloudScans[i]->points[j + 3].z +
                     laserCloudScans[i]->points[j + 4].z +
                     laserCloudScans[i]->points[j + 5].z;
      IdValue distance(j, diffX * diffX + diffY * diffY + diffZ * diffZ);
      cloudCurvature.push_back(distance);
    }

    for (int j = 0; j < 6; j++) {
      int sector_length = (int)(total_points / 6);
      int sector_start = sector_length * j;
      int sector_end = sector_length * (j + 1) - 1;
      if (j == 5) {
        sector_end = total_points - 1;
      }
      std::vector<IdValue> subCloudCurvature(
          cloudCurvature.begin() + sector_start,
          cloudCurvature.begin() + sector_end);

      extractSector(laserCloudScans[i], subCloudCurvature);
    }
  }

  // auto feats_less_flat_ds = PointCloudXYZI::Ptr(new PointCloudXYZI);
  pcl::VoxelGrid<PointXYZI> downSizeFilter;
  downSizeFilter.setInputCloud(feats_less_flat_raw_);
  downSizeFilter.setLeafSize(0.2, 0.2, 0.2);
  downSizeFilter.filter(*feats_less_flat_);
  // (*feats_less_flat_) += (*feats_less_flat_ds);
}

void FeatureExtractor::extractSector(
    const pcl::PointCloud<pcl::PointXYZI>::Ptr &pc_in,
    std::vector<IdValue> &cloudCurvature) {

  std::sort(
      cloudCurvature.begin(), cloudCurvature.end(),
      [](const IdValue &a, const IdValue &b) { return a.value < b.value; });

  int largestPickedNum = 0;
  std::vector<int> picked_points;
  int point_info_count = 0;
  for (int i = cloudCurvature.size() - 1; i >= 0; i--) {
    int ind = cloudCurvature[i].id;
    if (std::find(picked_points.begin(), picked_points.end(), ind) ==
        picked_points.end()) {
      if (cloudCurvature[i].value < 0.1) {
        break;
      }

      largestPickedNum++;
      picked_points.push_back(ind);

      if (largestPickedNum <= 2) {
        // cloudLabel[ind] = 2;
        feats_sharp_->push_back(pc_in->points[ind]);
        feats_less_sharp_->push_back(pc_in->points[ind]);
      } else if (largestPickedNum <= 20) {
        feats_less_sharp_->push_back(pc_in->points[ind]);
        // pc_out_edge->push_back(pc_in->points[ind]);
        point_info_count++;
      } else {
        break;
      }

      for (int k = 1; k <= 5; k++) {
        double diffX = pc_in->points[ind + k].x - pc_in->points[ind + k - 1].x;
        double diffY = pc_in->points[ind + k].y - pc_in->points[ind + k - 1].y;
        double diffZ = pc_in->points[ind + k].z - pc_in->points[ind + k - 1].z;
        if (diffX * diffX + diffY * diffY + diffZ * diffZ > 0.05) {
          break;
        }
        picked_points.push_back(ind + k);
      }
      for (int k = -1; k >= -5; k--) {
        double diffX = pc_in->points[ind + k].x - pc_in->points[ind + k + 1].x;
        double diffY = pc_in->points[ind + k].y - pc_in->points[ind + k + 1].y;
        double diffZ = pc_in->points[ind + k].z - pc_in->points[ind + k + 1].z;
        if (diffX * diffX + diffY * diffY + diffZ * diffZ > 0.05) {
          break;
        }
        picked_points.push_back(ind + k);
      }
    }
  }

  // for (int i = cloudCurvature.size() - 1; i >= 0; i--) {
  //   int ind = cloudCurvature[i].id;
  //   if (std::find(picked_points.begin(), picked_points.end(), ind) ==
  //       picked_points.end()) {
  //     if (cloudCurvature[i].value <= 0.1) {
  //       largestPickedNum++;
  //       picked_points.push_back(ind);

  //       if (largestPickedNum <= 2) {
  //         // cloudLabel[ind] = 2;
  //         feats_sharp_->push_back(pc_in->points[ind]);
  //         feats_less_sharp_->push_back(pc_in->points[ind]);
  //       } else if (largestPickedNum <= 20) {
  //         feats_less_sharp_->push_back(pc_in->points[ind]);
  //         // pc_out_edge->push_back(pc_in->points[ind]);
  //         point_info_count++;
  //       } else {
  //         break;
  //       }

  //       for (int k = 1; k <= 5; k++) {
  //         double diffX =
  //             pc_in->points[ind + k].x - pc_in->points[ind + k - 1].x;
  //         double diffY =
  //             pc_in->points[ind + k].y - pc_in->points[ind + k - 1].y;
  //         double diffZ =
  //             pc_in->points[ind + k].z - pc_in->points[ind + k - 1].z;
  //         if (diffX * diffX + diffY * diffY + diffZ * diffZ > 0.05) {
  //           break;
  //         }
  //         picked_points.push_back(ind + k);
  //       }
  //       for (int k = -1; k >= -5; k--) {
  //         double diffX =
  //             pc_in->points[ind + k].x - pc_in->points[ind + k + 1].x;
  //         double diffY =
  //             pc_in->points[ind + k].y - pc_in->points[ind + k + 1].y;
  //         double diffZ =
  //             pc_in->points[ind + k].z - pc_in->points[ind + k + 1].z;
  //         if (diffX * diffX + diffY * diffY + diffZ * diffZ > 0.05) {
  //           break;
  //         }
  //         picked_points.push_back(ind + k);
  //       }
  //     }
  //   }
  // }
  // find flat points
  // point_info_count =0;
  // int smallestPickedNum = 0;

  // for (int i = 0; i <= (int)cloudCurvature.size()-1; i++)
  // {
  //     int ind = cloudCurvature[i].id;

  //     if( std::find(picked_points.begin(), picked_points.end(),
  //     ind)==picked_points.end()){
  //         if(cloudCurvature[i].value > 0.1){
  //             //ROS_WARN("extracted feature not qualified, please check
  //             lidar"); break;
  //         }
  //         smallestPickedNum++;
  //         picked_points.push_back(ind);

  //         if(smallestPickedNum <= 4){
  //             //find all points
  //             pc_surf_flat->push_back(pc_in->points[ind]);
  //             pc_surf_lessFlat->push_back(pc_in->points[ind]);
  //             point_info_count++;
  //         }
  //         else{
  //             break;
  //         }

  //         for(int k=1;k<=5;k++){
  //             double diffX = pc_in->points[ind + k].x - pc_in->points[ind + k
  //             - 1].x; double diffY = pc_in->points[ind + k].y -
  //             pc_in->points[ind + k - 1].y; double diffZ = pc_in->points[ind
  //             + k].z - pc_in->points[ind + k - 1].z; if (diffX * diffX +
  //             diffY * diffY + diffZ * diffZ > 0.05){
  //                 break;
  //             }
  //             picked_points.push_back(ind+k);
  //         }
  //         for(int k=-1;k>=-5;k--){
  //             double diffX = pc_in->points[ind + k].x - pc_in->points[ind + k
  //             + 1].x; double diffY = pc_in->points[ind + k].y -
  //             pc_in->points[ind + k + 1].y; double diffZ = pc_in->points[ind
  //             + k].z - pc_in->points[ind + k + 1].z; if (diffX * diffX +
  //             diffY * diffY + diffZ * diffZ > 0.05){
  //                 break;
  //             }
  //             picked_points.push_back(ind+k);
  //         }

  //     }
  // }

  int smallestPickedNum = 0;
  for (int i = 0; i <= (int)cloudCurvature.size() - 1; i++) {
    int ind = cloudCurvature[i].id;
    if (std::find(picked_points.begin(), picked_points.end(), ind) ==
        picked_points.end()) {
      //   feats_less_flat_->push_back(pc_in->points[ind]);
      if (cloudCurvature[i].value <= 0.1) {
        smallestPickedNum++;
        picked_points.push_back(ind);
        feats_less_flat_raw_->push_back(pc_in->points[ind]);
        feats_flat_->push_back(pc_in->points[ind]);

        if (smallestPickedNum >= 4) {
          break;
        }
        // else if (smallestPickedNum <= 20) {
        //   feats_less_sharp_->push_back(pc_in->points[ind]);
        //   // pc_out_edge->push_back(pc_in->points[ind]);
        //   point_info_count++;
        // } else {
        //   break;
        // }

        for (int k = 1; k <= 5; k++) {
          double diffX =
              pc_in->points[ind + k].x - pc_in->points[ind + k - 1].x;
          double diffY =
              pc_in->points[ind + k].y - pc_in->points[ind + k - 1].y;
          double diffZ =
              pc_in->points[ind + k].z - pc_in->points[ind + k - 1].z;
          if (diffX * diffX + diffY * diffY + diffZ * diffZ > 0.05) {
            break;
          }
          picked_points.push_back(ind + k);
        }
        for (int k = -1; k >= -5; k--) {
          double diffX =
              pc_in->points[ind + k].x - pc_in->points[ind + k + 1].x;
          double diffY =
              pc_in->points[ind + k].y - pc_in->points[ind + k + 1].y;
          double diffZ =
              pc_in->points[ind + k].z - pc_in->points[ind + k + 1].z;
          if (diffX * diffX + diffY * diffY + diffZ * diffZ > 0.05) {
            break;
          }
          picked_points.push_back(ind + k);
        }
      }
      //   pc_out_surf->push_back();
    }
  }

  for (int i = 0; i <= (int)cloudCurvature.size() - 1; i++) {
    int ind = cloudCurvature[i].id;
    if (std::find(picked_points.begin(), picked_points.end(), ind) ==
        picked_points.end()) {
      feats_less_flat_raw_->push_back(pc_in->points[ind]);
      //   pc_out_surf->push_back();
    }
  }

  // surfPointsLessFlat += surfPointsLessFlatScanDS;
}

} // namespace xreg
