
#include "xreg/optimization/factors_pr.h"

#include "xreg/common/common.h"

#include "xreg/utils/math_utils.h"

namespace xreg {
using namespace std;

EigenFactorTrans2::EigenFactorTrans2(const Gaussian3d *info, const Vector3d &mu,
                                     const double s)
    : scalar_(s), info_map_(info), mu_(mu) {}

bool EigenFactorTrans2::Evaluate(double const *const *parameters,
                                 double *residuals, double **jacobians) const {

  Eigen::Map<const Vector3d> rot_log(parameters[0]);
  Eigen::Map<const Vector3d> t(parameters[1]);
  const auto &rot = SO3::exp(rot_log);

  const RowVector3d &nhat = info_map_->rot_mat.col(0).transpose();
  const Vector3d &mu_tgt = info_map_->mean;

  if (residuals) {
    residuals[0] = scalar_ * nhat.dot((rot * mu_ + t) - mu_tgt);
    if (isnan(residuals[0])) {
      WARNLOG("fuck");
      exit(0);
    }
  }

  if (jacobians) {
    if (jacobians[0]) {
      Eigen::Map<RowVector3d> jacob(jacobians[0]);
      jacob = -nhat * rot.matrix() * MathUtils::skew(mu_);
      jacob *= scalar_;
    }

    if (jacobians[1]) {
      Eigen::Map<RowVector3d> jacob(jacobians[1]);
      jacob = nhat * scalar_;
    }
  }

  return true;
}

EigenFactorRotAxis::EigenFactorRotAxis(const Gaussian3d *cell,
                                       const Vector3d &axis, const double s)
    : scalar_(s), info_map_(cell), axis_(axis) {}

bool EigenFactorRotAxis::Evaluate(double const *const *parameters,
                                  double *residuals, double **jacobians) const {

  Eigen::Map<const Vector3d> rot_log(parameters[0]);
  const auto &rot = SO3::exp(rot_log);

  const RowVector3d &nhat = info_map_->rot_mat.col(0).transpose();

  if (residuals) {
    residuals[0] = nhat.dot(rot * axis_) * scalar_;
    if (isnan(residuals[0])) {
      WARNLOG("fuck");
      cout << axis_.transpose() << endl;
      cout << scalar_ << endl;
      cout << nhat << endl;
      exit(0);
    }
  }

  if (jacobians) {
    if (jacobians[0]) {
      Eigen::Map<RowVector3d> jacob(jacobians[0]);
      jacob = -scalar_ * (nhat * rot.matrix() * MathUtils::skew(axis_));
    }
  }

  return true;
}

void SurfRegXXEvalCallBack::PrepareForEvaluation(bool evaluate_jacobians,
                                                 bool new_evaluation_point) {
  // prepare for evaluation
  // INFOLOG("callback for eval, eval_jacob: {}, eva_point: {}...",
  //         evaluate_jacobians, new_evaluation_point);

  if (new_evaluation_point) {
    for (size_t i = 0; i < frames_.size(); i++) {
      auto &&frame = frames_[i];
      double *pr = param_rot_[i];
      double *pt = param_trans_[i];

      Eigen::Map<const Vector3d> rot_log(pr);
      Eigen::Map<const Vector3d> t(pt);
      frame->setTwb(SE3(SO3::exp(rot_log), t));
    }

    // for (auto &&info : record_)
    for (size_t i = 0; i < cells_.size(); i++) {
      // auto &&plane = info.first;
      auto &&plane = cells_[i];
      plane->updateInfoAll(false);
    }
    // INFOLOG("callback exit..");
  }
}

void SurfRegXXEvalCallBack::addPlane(Cell *plane) { cells_.push_back(plane); }

void SurfRegXXEvalCallBack::addFrame(Frame *f, double *rot, double *trans) {
  frames_.push_back(f);
  param_rot_.push_back(rot);
  param_trans_.push_back(trans);
}

} // namespace xreg