#include "xreg/test_scan2scan.h"

#include <chrono>
#include <fstream>

#include <pcl/common/transforms.h>
#include <pcl/filters/voxel_grid.h>
#include <pcl/io/pcd_io.h>

#include "xreg/estimation/registration_ba.h"
#include "xreg/estimation/registration_base.h"
#include "xreg/estimation/registration_fpfh.h"
// #include "xreg/estimation/registration_gicp.h"
#include "xreg/estimation/registration_icp_pt2pl.h"
// #include "xreg/estimation/registration_icp_pt2pt.h"
#include "xreg/estimation/registration_ndt.h"
// #include "xreg/estimation/registration_vgicp.h"

#include "xreg/common/config.h"

namespace xreg {

using namespace std;

TestScan2Scan::TestScan2Scan(ros::NodeHandle nh) : nh_(nh) {

  // viewer_ = std::unique_ptr<Visualizer>(new Visualizer(nh_));
  // thread_viz_ =
  //     unique_ptr<thread>(new std::thread(&Visualizer::spin, viewer_.get()));

  initParameters(nh_);

  loadFrames("/home/hyhuang/Downloads/" + common::dataset);
}

void TestScan2Scan::spin() {

  // visualizeIdx();

  registration_ = new RegistrationBA;

  // registration_->setInputFrames(frames_);
  // registration_->initialize();

  size_t num_frames = frames_.size();
  std::vector<double> errors;
  for (size_t i = 0; i < frames_.size() - 1; i++) {
    auto start = std::chrono::steady_clock::now();
    auto frame_tgt = frames_[i];
    auto frame_src = frames_[i + 1];
    registration_->scanToScan(frame_tgt, frame_src);
    // re
    auto end = std::chrono::steady_clock::now();
    auto duration =
        std::chrono::duration_cast<std::chrono::milliseconds>(end - start);
    SE3 Test = frame_tgt->getTbw() * frame_src->getTwb();
    SE3 Tgt = frame_tgt->Tgt_.inverse() * frame_src->Tgt_;
    SE3 Terr = Tgt.inverse() * Test;
    errors.push_back(Terr.translation().norm());

    // Test
  }

  double err =
      std::accumulate(errors.begin(), errors.end(), 0.0) / errors.size();
  INFOLOG("mean error: {}", err);
  INFOLOG("max error: {}", *std::max_element(errors.begin(), errors.end()));
  INFOLOG("min error: {}", *std::min_element(errors.begin(), errors.end()));
  for (size_t i = 0; i < errors.size(); i++) {
    INFOLOG("idx: {}, error: {}", i, errors[i]);
  }

  // std::cout << "elapsed time: " << duration.count() << " ms\n";

  // summarize(duration.count());
}

void TestScan2Scan::summarize(double time) {
  // x
  // double err_rot = 0, err_trans = 0, err_rel = 0;
  // for (size_t i = 1; i < frames_.size(); i++) {
  //   auto frame_src = frames_[i];
  //   auto frame_tgt = frames_[i - 1];

  //   SE3 deltaT_gt = frame_tgt->Tgt_.inverse() * frame_src->Tgt_;
  //   SE3 deltaT = frame_tgt->getTbw() * frame_src->getTwb();
  //   SE3 deltaT_err = (deltaT_gt.inverse() * deltaT);

  //   double err_rel_frame = deltaT_err.translation().norm();
  //   err_rel += err_rel_frame;

  //   SE3 Terr = frame_tgt->Tgt_.inverse() * frame_tgt->getTwb();

  //   double err_trans_frame = Terr.translation().norm();
  //   err_trans += err_trans_frame;

  //   double err_rot_frame = Terr.so3().log().norm() / M_PI * 180.0;
  //   err_rot += err_rot_frame;
  // }

  // double time_frame = time;
  // err_rel /= (frames_.size() - 1);
  // time_frame /= (frames_.size() - 1);
  // err_trans /= (frames_.size() - 1);
  // err_rot /= (frames_.size() - 1);

  // WARNLOG("err_rel: {} , err_t: {}, err_rot: {}, time: {}", err_rel,
  // err_trans,
  //         err_rot, time_frame);
  // WARNLOG("{}, {}, {}, {}", err_rel, err_trans, err_rot, time_frame);
}

void TestScan2Scan::shutdown() {
  // stop viewer
  if (viewer_) {
    viewer_->stop();

    if (thread_viz_) {
      thread_viz_->join();
    }
  }
}

void TestScan2Scan::loadFrames(const std::string &file_path) {
  // load csv
  {
    std::string csv_file = file_path + "/icpList.csv";
    ifstream ifs(csv_file);

    if (!ifs.is_open()) {
      ERRLOG("file not opened: {}", csv_file);
    }

    string line;
    // header
    getline(ifs, line);

    // data
    while (!ifs.eof() && std::getline(ifs, line)) {
      std::stringstream ss(line);
      string fname;
      ss >> fname;
      int colIdx = 0;
      double val;
      vector<double> vals;
      while (ss >> val) {
        // If the next token is a comma, ignore it and move on
        if (ss.peek() == ',')
          ss.ignore();

        // INFOLOG("val: {}", val);
        vals.push_back(val);
      }
      //   exit(0);
      Matrix3d rot_mat;
      rot_mat << vals[0], vals[1], vals[2], vals[4], vals[5], vals[6], vals[8],
          vals[9], vals[10];

      Quaterniond rot(rot_mat);
      rot.normalize();

      Vector3d trans;
      trans << vals[3], vals[7], vals[11];

      auto frame = Frame::Ptr(new Frame);
      frames_.push_back(frame);
      frame->Tgt_ = SE3(rot, trans);
    }
  }
  INFOLOG("num_frames: {}", frames_.size());
  INFOLOG("last t: {}", frames_.back()->Tgt_.translation().transpose());

  for (size_t i = 0; i < frames_.size(); i++) {
    string pcd_file = file_path + "/pcd/" + to_string(i) + ".pcd";

    auto cloud = PointCloudXYZ::Ptr(new PointCloudXYZ);
    pcl::io::loadPCDFile(pcd_file, *cloud);
    frames_[i]->cloud_ = cloud;
  }

  //   visualize();
}

inline Vector3i makeJet3B(float id) {
  if (id <= 0)
    return Vector3i(128, 0, 0);
  if (id >= 1)
    return Vector3i(0, 0, 128);

  int icP = (id * 8);
  float ifP = (id * 8) - icP;

  if (icP == 0)
    return Vector3i(255 * (0.5 + 0.5 * ifP), 0, 0);
  if (icP == 1)
    return Vector3i(255, 255 * (0.5 * ifP), 0);
  if (icP == 2)
    return Vector3i(255, 255 * (0.5 + 0.5 * ifP), 0);
  if (icP == 3)
    return Vector3i(255 * (1 - 0.5 * ifP), 255, 255 * (0.5 * ifP));
  if (icP == 4)
    return Vector3i(255 * (0.5 - 0.5 * ifP), 255, 255 * (0.5 + 0.5 * ifP));
  if (icP == 5)
    return Vector3i(0, 255 * (1 - 0.5 * ifP), 255);
  if (icP == 6)
    return Vector3i(0, 255 * (0.5 - 0.5 * ifP), 255);
  if (icP == 7)
    return Vector3i(0, 0, 255 * (1 - 0.5 * ifP));
  return Vector3i(255, 255, 255);
}

void TestScan2Scan::visualizeIdx() {

  {
    // PointCloudXYZ::Ptr cloud_map_raw(new PointCloudXYZ);
    // downsample(cloud_map_raw, cloud_map, 0.05);

    pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_map(
        new pcl::PointCloud<pcl::PointXYZRGB>);
    for (size_t i = 0; i < frames_.size(); i++) {
      auto frame = frames_[i];
      // viewer_->addCloud("cloud" + to_string(i), "frame" + to_string(i),
      //                   frame->cloud_);

      // viewer_->setTransform(frame->Tgt_, "frame" + to_string(i));

      Vector3i c = makeJet3B((float)i / frames_.size());

      PointCloudXYZ::Ptr cloud_tmp(new PointCloudXYZ);
      pcl::transformPointCloud(*frame->cloud_, *cloud_tmp,
                               frame->Tgt_.matrix().cast<float>());
      pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_color(
          new pcl::PointCloud<pcl::PointXYZRGB>);

      for (auto &&pt : cloud_tmp->points) {
        pcl::PointXYZRGB ptc(c.x(), c.y(), c.z());
        ptc.x = pt.x;
        ptc.y = pt.y;
        ptc.z = pt.z;

        cloud_color->push_back(ptc);
      }

      (*cloud_map) += *cloud_color;
    }

    viewer_->addCloud("cloud_map_idx", "map", *cloud_map);

    // PointCloudXYZ::Ptr cloud_map(new PointCloudXYZ);
    // downsample(cloud_map_raw, cloud_map, 0.05);

    // viewer_->addCloud("cloud_map_gt", "map", cloud_map);
  }
}

void TestScan2Scan::visualize() {

  {
    PointCloudXYZ::Ptr cloud_map_raw(new PointCloudXYZ);
    for (size_t i = 0; i < frames_.size(); i++) {
      auto frame = frames_[i];
      // viewer_->addCloud("cloud" + to_string(i), "frame" + to_string(i),
      //                   frame->cloud_);

      // viewer_->setTransform(frame->Tgt_, "frame" + to_string(i));

      PointCloudXYZ::Ptr cloud_tmp(new PointCloudXYZ);
      pcl::transformPointCloud(*frame->cloud_, *cloud_tmp,
                               frame->Tgt_.matrix().cast<float>());

      (*cloud_map_raw) += *cloud_tmp;
    }

    viewer_->addCloud("cloud_map_gt", "map", *cloud_map_raw);

    // PointCloudXYZ::Ptr cloud_map(new PointCloudXYZ);
    // downsample(cloud_map_raw, cloud_map, 0.05);

    // viewer_->addCloud("cloud_map_gt", "map", cloud_map);
  }

  {
    PointCloudXYZ::Ptr cloud_map_raw(new PointCloudXYZ);
    for (size_t i = 0; i < frames_.size(); i++) {
      auto frame = frames_[i];

      PointCloudXYZ::Ptr cloud_tmp(new PointCloudXYZ);
      pcl::transformPointCloud(*frame->cloud_, *cloud_tmp,
                               frame->getTwb().matrix().cast<float>());

      (*cloud_map_raw) += *cloud_tmp;
    }
    viewer_->addCloud("cloud_map_est", "map", *cloud_map_raw);

    // PointCloudXYZ::Ptr cloud_map(new PointCloudXYZ);
    // downsample(cloud_map_raw, cloud_map, 0.05);

    // viewer_->addCloud("cloud_map_est", "map", cloud_map);
  }
}

void TestScan2Scan::downsample(PointCloudXYZ::Ptr cloud,
                               PointCloudXYZ::Ptr cloud_filtered, double size) {
  pcl::VoxelGrid<PointXYZ> filter;
  filter.setLeafSize(size, size, size);
  filter.setInputCloud(cloud);
  filter.filter(*cloud_filtered);
}

} // namespace xreg
