#include "xreg/types/frame.hpp"

namespace xreg {
using namespace std;

double Frame::countOverlap(Frame *frame) {
  // segments_.

  unordered_set<Cell *> cells;
  for (auto &&seg : segments_) {
    cells.insert(seg->cell_);
  }

  int count = 0;
  for (auto &&seg : frame->segments_) {
    if (cells.count(seg->cell_)) {
      count++;
    }
  }

  return count * 1.0 / frame->segments_.size();
}
//
// Frame::Frame(PointCloudXYZ::Ptr cloud, SE3 Twb)
bounds_t bounds_t::get_intersection_bbx(bounds_t &bbx_1, bounds_t &bbx_2,
                                        // bounds_t &bbx_intersection,
                                        float bbx_boundary_pad) {
  bounds_t bbx_intersection;
  bbx_intersection.min_x = max(bbx_1.min_x, bbx_2.min_x) - bbx_boundary_pad;
  bbx_intersection.min_y = max(bbx_1.min_y, bbx_2.min_y) - bbx_boundary_pad;
  bbx_intersection.min_z = max(bbx_1.min_z, bbx_2.min_z) - bbx_boundary_pad;
  bbx_intersection.max_x = min(bbx_1.max_x, bbx_2.max_x) + bbx_boundary_pad;
  bbx_intersection.max_y = min(bbx_1.max_y, bbx_2.max_y) + bbx_boundary_pad;
  bbx_intersection.max_z = min(bbx_1.max_z, bbx_2.max_z) + bbx_boundary_pad;

  return bbx_intersection;
}

double bounds_t::get_intersection_ratio(bounds_t &bbox) {
  // get
  auto box_inter = get_intersection_bbx(*this, bbox);
  auto v0 = box_inter.volume();
  auto v1 = this->volume();
  auto v2 = bbox.volume();

  double r0 = (v0 / v1);
  double r1 = (v0 / v2);
  // return max(r0, r1);
  return (r0 + r1) / 2.0;
}

bounds_t Frame::getBBox() const { return bbox_; }

bounds_t Frame::getTransformedBBox(const SE3 &Twb) const {
  // Twb
  Vector3d pt_min = Twb * Vector3d(bbox_.min_x, bbox_.min_y, bbox_.min_z);
  Vector3d pt_max = Twb * Vector3d(bbox_.max_x, bbox_.max_y, bbox_.max_z);

  bounds_t bbox;
  bbox.min_x = pt_min.x();
  bbox.min_y = pt_min.y();
  bbox.min_z = pt_min.z();

  bbox.max_x = pt_max.x();
  bbox.max_y = pt_max.y();
  bbox.max_z = pt_max.z();

  return bbox;
}

// bounds_t Frame::getBBoxGlobal() const {
//     bounds_t bbox;
//     Vecto3d min_pt = Tgt_ * bbox_
// }
void Frame::setBBox(const PointXYZ &pt_min, const PointXYZ &pt_max) {
  bbox_.min_x = pt_min.x;
  bbox_.min_y = pt_min.y;
  bbox_.min_z = pt_min.z;

  bbox_.max_x = pt_max.x;
  bbox_.max_y = pt_max.y;
  bbox_.max_z = pt_max.z;
}

Vector6d Frame::getJacobian() {
  //   jacobian_.setZero();
  Vector6d jacobian_ = Vector6d::Zero();

  for (auto &&seg : segments_) {
    if (seg->cell_->frames_.size() < 2) {
      continue;
    }

    Matrix4d dQ;
    Matrix4d Q = seg->Q_;
    Vector4d pi = seg->cell_->pi_;
    // cout << Q << endl;
    // cout << pi << endl;
    // Matrix4d Q;
    // Vector4d pi;

    Vector6d jacobian;
    // calculate dQ/dxi for the submatrix S
    // Mat4 dQ, &Q = matrixQ_[t];

    // dQ / d xi(3) = [q4
    //                 0
    //                 0
    //                 0]
    dQ.setZero();
    dQ.row(0) << Q.row(3);
    dQ += dQ.transpose().eval();
    jacobian(0) = pi.dot(dQ * pi);

    // dQ / d xi(4) = [0
    //                 q4
    //                 0
    //                 0]
    dQ.setZero();
    dQ.row(1) << Q.row(3);
    dQ += dQ.transpose().eval();
    jacobian(1) = pi.dot(dQ * pi);

    // dQ / d xi(5) = [0
    //                 0
    //                 q4
    //                 0]
    dQ.setZero();
    dQ.row(2) << Q.row(3);
    dQ += dQ.transpose().eval();
    jacobian(2) = pi.dot(dQ * pi);

    // dQ / d xi(0) = [0
    //               -q3
    //                q2
    //                 0]
    dQ.setZero();
    dQ.row(1) << -Q.row(2);
    dQ.row(2) << Q.row(1);
    dQ += dQ.transpose().eval();
    jacobian(3) = pi.dot(dQ * pi);

    // dQ / d xi(1) = [q3
    //                 0
    //                -q1
    //                 0]
    dQ.setZero();
    dQ.row(0) << Q.row(2);
    dQ.row(2) << -Q.row(0);
    dQ += dQ.transpose().eval();
    jacobian(4) = pi.dot(dQ * pi);

    // dQ / d xi(2) = [-q2
    //                 q1
    //                 0
    //                 0]
    dQ.setZero();
    dQ.row(0) << -Q.row(1);
    dQ.row(1) << Q.row(0);
    dQ += dQ.transpose().eval();
    jacobian(5) = pi.dot(dQ * pi);

    jacobian_ += jacobian;
  }

  return jacobian_;
}

int Frame::getNumPoints() {
  int num = 0;
  for (auto &&p : segments_) {
    if (p->cell_->frames_.size() < 2) {
      continue;
    }
    // num += p->pts_.size();
    num += p->num_;
  }

  return num;
}

}  // namespace xreg
