// #include<pointcloud_vrml/pointcloud_utils.h>
#include "xreg/types/octree.h"

#include "xreg/types/cell.h"

namespace xreg {

using namespace std;

unsigned int OctreeNode::MAX_DEPTH;
/// @param number of points after which to split cell
unsigned int OctreeNode::MAX_POINTS;
/// @param at this level do not split any more
double OctreeNode::SMALL_CELL_SIZE;
/// @param split tree up to this size before allocating a cell
double OctreeNode::BIG_CELL_SIZE;

void OctreeNode::setParameters(double _BIG_CELL_SIZE, double _SMALL_CELL_SIZE,
                               int _MAX_POINTS, int _MAX_DEPTH) {

  // defaults
  MAX_POINTS = _MAX_POINTS;
  MAX_DEPTH = _MAX_DEPTH;
  BIG_CELL_SIZE = _BIG_CELL_SIZE;
  SMALL_CELL_SIZE = _SMALL_CELL_SIZE;

  // if (!this->is_leaf_) {
  //   for (unsigned int it = 0; it < 8; it++) {
  //     if (this->children_[it] != nullptr) {
  //       this->children_[it]->setParameters(_BIG_CELL_SIZE, _SMALL_CELL_SIZE,
  //                                          _MAX_POINTS, _MAX_DEPTH);
  //     }
  //   }
  // }
}

/** empty constructor
 */
OctreeNode::OctreeNode() {

  this->parent_ = nullptr;
  this->is_leaf_ = true;
  this->depth_ = 0;
  this->size_ = 0;
  this->center_.setZero();

  for (unsigned int it = 0; it < 8; it++) {
    this->children_[it] = nullptr;
  }

  // myCell_ = nullptr;
  // leafsCached_ = false;

  // random color
  color_ << (Vector3d::Random() + Vector3d::Ones()) / 2.0;

  // this->setParameters();
}

/** parametrized constructor
 */
// template <typename PointT>
OctreeNode::OctreeNode(Vector3d center, double xsize,
                       // double ysize, double zsize,
                       //  NDTCell<PointT> *type,
                       OctreeNode *_parent, unsigned int _depth) {

  parent_ = _parent;
  is_leaf_ = true;
  depth_ = _depth;
  size_ = xsize;
  for (unsigned int it = 0; it < 8; it++) {
    children_[it] = nullptr;
  }

  center_ = center;

  // random color
  color_ << (Vector3d::Random() + Vector3d::Ones()) / 2.0;
  //   NDTCell<PointT> *tmp = dynamic_cast<NDTCell<PointT> *>(type->clone());
  //   if (tmp == NULL) {
  //     // ERR("dynamic cast of cell failed!!\n");
  //     return;
  //   }

  //   tmp->setCenter(center);
  //   tmp->setDimensions(xsize, ysize, zsize);
  //   tmp->points_.clear();

  //   myCell_ = tmp;
  // leafsCached_ = false;
  // this->setParameters();
}

/** destructor, deletes all pointers starting from parent and going down
 */
// template <typename PointT>
OctreeNode::~OctreeNode() {
  if (!is_leaf_) {
    for (unsigned int it = 0; it < 8; it++) {
      if (children_[it] != nullptr) {
        delete children_[it]; // calls destructor of child, so we are ok
        children_[it] = nullptr;
      }
    }
  } else {
    for (auto &&segment : cell_->segments_) {
      delete segment;
    }
    delete cell_;
  }

  // delete myCell_;
}

// template <typename PointT>

/** returns the child index that should hold the point.
    \note the point is not necessarily within the child's boundaries,
    but is guaranteed to be in the same quadrant (octant?)
  */
// template <typename PointT>
size_t OctreeNode::getIndexForPoint(const Vector3d &pt) const {
  /** index table: (convenient computations)
   *	id  x	y   z
   * 0   +	+   +
   * 1   +	+   -
   * 2   +	-   -
   * 3   +	-   +
   * 4   -	+   +
   * 5   -	+   -
   * 6   -	-   -
   * 7   -	-   +
   */
  size_t ret = 0;
  if (pt.x() < center_.x()) {
    ret += 4;
  }
  if (pt.y() < center_.y()) {
    ret += 2;
  }
  if (ret % 4 == 0) {
    if (pt.z() < center_.z()) {
      ret += 1;
    }
  } else {
    if (pt.z() > center_.z()) {
      ret += 1;
    }
  }

  return ret;
}

/** finds all children cells that are located at current leafs
 */
void OctreeNode::computeLeafCells() {
  throw std::runtime_error("not implemented");
}

/** adds a point to the index. Iterates down the tree and if necessary
  creates new leafs and splits current ones.
  \note at the moment root is not grown in case of points outside!
  */
// template <typename PointT>
// Cell<PointT> *OctTree<PointT>::addPoint(const PointT &point_c) {
std::tuple<OctreeNode *, Cell *, Segment *>
OctreeNode::addPoint(Frame *frame, const Vector3d &Tpt, const Vector3d &pt) {

  throw std::runtime_error("not implemented");

  // // leafsCached_ = false;
  // if (cell_) {
  //   auto segment = cell_->getSegment(frame);
  //   segment->cachePoint(Tpt, pt);
  //   return make_tuple(this, cell_, segment);
  // } else if (is_leaf_) {
  //   cell_ = new Cell;
  //   auto segment = cell_->getSegment(frame);
  //   CHECK_NOTNULL(segment);
  //   segment->addPoint(Tpt, pt);

  //   return make_tuple(this, cell_, segment);
  // } else {
  //   size_t ind = getIndexForPoint(Tpt);
  //   if (!children_[ind]) {
  //     expandNode(ind);
  //   }

  //   return children_[ind]->addPoint(frame, Tpt, pt);
  // }
}

/** adds a point to the index. Iterates down the tree and if necessary
  creates new leafs and splits current ones.
  \note at the moment root is not grown in case of points outside!
  */
// template <typename PointT>
// Cell<PointT> *OctTree<PointT>::addPoint(const PointT &point_c) {
// OctreeNode *OctreeNode::addPoint(const Vector3d &point) {

//   leafsCached_ = false;
//   if (this->is_leaf_) {
//     // if reach max_depth or number of points is not large, push
//     // otherwise, expand nodes
//     if (depth_ == MAX_DEPTH || points_.size() < MAX_POINTS) {
//       points_.push_back(point);
//       return this;
//     } else {

//       // add current points && branch leaves if necessary
//       for (size_t jt = 0; jt < points_.size(); jt++) {
//         const Vector3d &pt = points_[jt];
//         size_t ind = getIndexForPoint(pt);

//         if (!children_[ind]) {
//           expandNode(ind);
//         }
//         children_[ind]->addPoint(pt);
//       }
//       points_.clear();

//       // finally add the new point
//       this->is_leaf_ = false;

//       size_t ind = getIndexForPoint(point);
//       if (!children_[ind]) {
//         expandNode(ind);
//       }

//       return children_[ind]->addPoint(point);
//     }

//   } else {
//     // pass down to correct child
//     size_t ind = getIndexForPoint(point);

//     // CHECK_NOTNULL(children_[ind]);
//     if (!children_[ind]) {
//       expandNode(ind);
//     }

//     return children_[ind]->addPoint(point);
//   }
// }

void OctreeNode::getCells(std::vector<Cell *> &cells) {
  if (cell_) {
    cells.push_back(cell_);
  }

  for (auto &&child : children_) {
    if (child) {
      child->getCells(cells);
    }
  }
}

// bool OctreeNode::divideCell() {
//   if (!cell_ || depth_ == MAX_DEPTH) {
//     return false;
//   }

//   cell_->solveX();
//   if (!cell_->optable_) {
//     return true;
//   }

//   if (!cell_->checkX()) {
//     vector<Frame *> frames = cell_->frames_;
//     vector<Segment *> segments = cell_->segs_;

//     unordered_set<Segment *> segment_cache;

//     CHECK_EQ(frames.size(), segments.size());
//     for (size_t i = 0; i < segments.size(); i++) {
//       // segments.
//       auto frame = frames[i];
//       auto seg = segments[i];

//       // for (auto &&pt : seg->pts_)
//       for (size_t i = 0; i < seg->pts_.size(); i++) {
//         Vector3d &pt = seg->pts_[i];
//         Vector3d &Tpt = seg->Tpts_[i];

//         // pt->
//         size_t ind = getIndexForPoint(pt);
//         if (!children_[ind]) {
//           expandNode(ind);
//         }

//         auto res = children_[ind]->addPoint(frame, Tpt, pt);

//         auto new_seg = std::get<2>(res);
//         segment_cache.insert(new_seg);
//         new_seg->frame_ = frame;
//       }
//     }

//     this->is_leaf_ = false;
//     delete this->cell_;
//     this->cell_ = nullptr;

//     // NOTE: addPoint with frame can only retrieve single level
//     for (auto &&seg : segment_cache) {
//       seg->estimate2();
//       // seg->frame_ = frame;
//     }
//     for (auto &&child : children_) {
//       if (child) {
//         child->divideCell();
//       }
//     }
//   }
// }

OctreeNode *OctreeNode::expandNode(int ind) {
  if (ind == -1) {
    // idx
    // TODO: expand all the node
    throw std::runtime_error("not implemented");
  } else {
    if (!children_[ind]) {
      Vector3d cchild;

      // computes the center_ of the it'th child
      cchild[0] = center_.x() + pow(-1., ind / 4) * size_ / 4.;
      cchild[1] = center_.y() + pow(-1., ind / 2) * size_ / 4.;
      cchild[2] = center_.z() + pow(-1., (ind + 1) / 2) * size_ / 4.;

      children_[ind] = new OctreeNode(cchild, size_ / 2.0, this, depth_ + 1);
    }
  }
}

/** returns the cell that should hold the point
 */
// template <typename PointT>
OctreeNode *OctreeNode::getNodeForPoint(const Vector3d &point, int depth) {
  OctreeNode *pointLeaf = this->getLeafForPoint(point);
  return (pointLeaf == nullptr) ? nullptr : pointLeaf;
}

/** finds the leaf that should hold the point
 */
OctreeNode *OctreeNode::getLeafForPoint(const Vector3d &point,
                                        bool create_new) {
  if (this->is_leaf_) {
    // if (myCell_->isInside(point)) {
    //   return this;
    // }

    // TODO: check inside
    return this;
  } else {
    size_t ind = getIndexForPoint(point);

    if (children_[ind] != nullptr) {
      return children_[ind]->getLeafForPoint(point);
    } else if (create_new) {
      return expandNode(ind);
    }
  }

  return nullptr;
}

// void Cell::computeGaussian(const eigen_aligned_std_vector<Vector3d> &pts) {
//   num_ = pts.size();

//   // mean
//   Eigen::Vector3d meanSum_;
//   mean_ << 0, 0, 0;
//   for (auto &&pt : pts) {
//     mean_ += pt;
//   }
//   // meanSum_ = mean_;
//   mean_ /= num_;

//   // cov
//   Eigen::MatrixXd mp;
//   mp.resize(pts.size(), 3);
//   for (unsigned int i = 0; i < pts.size(); i++) {
//     mp.row(i) = pts[i] - mean_;
//   }
//   Eigen::Matrix3d covSum_ = mp.transpose() * mp;
//   cov_ = covSum_ / (num_ - 1);

//   // find the norm of plane
//   // Eigen::Vector3d norm = matA0.colPivHouseholderQr().solve(matB0);
//   // double negative_OA_dot_norm = 1 / norm.norm();
//   // norm.normalize();

//   Eigen::SelfAdjointEigenSolver<Eigen::Matrix3d> saes(cov_);
//   axes_ = saes.eigenvectors();
//   scales_ = saes.eigenvalues();

//   // Eigen::Vector3d unit_direction = saes.eigenvectors().col(2);
//   // Eigen::Vector3d curr_point(pc_in->points[i].x, pc_in->points[i].y,
//   //                            pc_in->points[i].z);
// }

// /** sets the cell factory type
//  */
// template <typename PointT>
// void OctTree<PointT>::setCellType(Cell<PointT> *type) {

//   myCell_ = dynamic_cast<NDTCell<PointT> *>(type->clone());
//   if (myCell_ == NULL) {
//     // cast failed, it's not a derivative of oct cell
//     myCell_ = new NDTCell<PointT>();
//   }
// }

/** iterator poining to the first leaf cell.
 * \note recomputes the vector when changes have occured!
 */
// template <typename PointT>
// typename SpatialIndex<PointT>::CellVectorItr OctTree<PointT>::begin() {
//   if (!leafsCached_) {
//     myLeafs_.clear();
//     computeLeafCells();
//   }
//   leafsCached_ = true;
//   return myLeafs_.begin();
// }

/** iterator poining at last leaf cell
 */
// template <typename PointT>
// typename SpatialIndex<PointT>::CellVectorItr OctTree<PointT>::end() {
//   if (!leafsCached_) {
//     myLeafs_.clear();
//     computeLeafCells();
//   }
//   leafsCached_ = true;
//   return myLeafs_.end();
// }

/** returns a new OctTree spatial index
 */
// template <typename PointT>
// SpatialIndex<PointT> *OctTree<PointT>::clone() const {
//   OctTree<PointT> *tr = new OctTree<PointT>();
//   tr->setParameters(BIG_CELL_SIZE, SMALL_CELL_SIZE, MAX_POINTS, MAX_DEPTH);
//   if (myCell_ != NULL) {
//     tr->setCellType(myCell_);
//   }
//   return tr;
// }

/** copies the spatial index
 */
// template <typename PointT> SpatialIndex<PointT> *OctTree<PointT>::copy()
// const {
//   OctTree<PointT> *tr;
//   if (myCell_ != NULL) {
//     PointT center = myCell_->getCenter();
//     double sx, sy, sz;
//     myCell_->getDimensions(sx, sy, sz);
//     tr = new OctTree<PointT>(center, sx, sy, sz, myCell_);
//   } else {
//     tr = new OctTree<PointT>();
//   }
//   tr->setParameters(BIG_CELL_SIZE, SMALL_CELL_SIZE, MAX_POINTS, MAX_DEPTH);
//   return tr;
// }

/** sets the center_.
  \note this is not going to re-compute cells that are already in the tree!
  */
// template <typename PointT>
// void OctTree<PointT>::setCenter(const double &cx, const double &cy,
//                                 const double &cz) {
//   if (myCell_ == NULL) {
//     return;
//   }
//   PointT center_;
//   center_.x = cx;
//   center_.y = cy;
//   center_.z = cz;

//   myCell_->setCenter(center_);
// }

/** sets the size
  \note this is not going to re-compute cells that are already in the tree!
  */
// template <typename PointT>
// void OctTree<PointT>::setSize(const double &sx, const double &sy,
//                               const double &sz) {
//   if (myCell_ == NULL) {
//     return;
//   }
//   myCell_->setDimensions(sx, sy, sz);
// }

// template <typename PointT>
// NDTCell<PointT> *OctTree<PointT>::getClosestNDTCell(const PointT &point) {
//   if (this->leaf_) {
//     // we have reached the bottom of the tree.
//     // if we have a valid ndt cell, return it.
//     if (myCell_->isInside(point)) {
//       NDTCell<PointT> *nd = dynamic_cast<NDTCell<PointT> *>(myCell_);
//       if (nd != NULL) {
//         if (nd->hasGaussian_) {
//           return nd;
//         }
//       }
//     }
//   }

//   // we go down the tree recursively
//   size_t ind = getIndexForPoint(point);
//   if (children_[ind] != NULL) {
//     return children_[ind]->getClosestNDTCell(point);
//   }

//   // the leaf we should be in is empty
//   // iterate through all leafs connected to current node, find closest ndt
//   cell OctTree<PointT> *my_parent = this->parent_; OctTree<PointT> *me =
//   this; typename SpatialIndex<PointT>::CellVectorItr it; NDTCell<PointT>
//   *closest = NULL, *temp = NULL; double minDist = INT_MAX, dist = INT_MAX;

//   while (true) {
//     it = me->begin();
//     while (it != me->end()) {
//       temp = dynamic_cast<NDTCell<PointT> *>(*it);
//       if (temp != NULL) {
//         if (temp->hasGaussian_) {
//           dist = perception_oru::geomDist<PointT>(temp->getCenter(), point);
//           if (dist < minDist) {
//             minDist = dist;
//             closest = temp;
//           }
//         }
//       }
//       it++;
//     }
//     if (closest != NULL) {
//       //	    cout<<"got it!\n";
//       break;
//     }
//     if (my_parent != NULL) {
//       me = my_parent;
//       my_parent = me->parent_;
//     } else {
//       // nothing more can be done...
//       break;
//     }
//   }

//   return closest;
// }

} // namespace xreg
