#include "xreg/types/voxel.h"

#include <tuple>

#include <pcl/common/transforms.h>

#include "xreg/common/config.h"

namespace xreg {

using namespace std;

Block::Block(double block_size, Vector3d origin)
    : block_size_(block_size),
      block_size_inv_(1.0 / block_size),
      origin_(origin) {
  VoxelIndex voxel_idx =
      getGridIndexFromPoint<VoxelIndex>(origin * block_size_inv_);
  idx_ = AnyIndexHash()(voxel_idx);

  initOctree();
}

void Block::initOctree() {
  Vector3d center = origin_;
  center += Vector3d(block_size_ / 2.0, block_size_ / 2.0, block_size_ / 2.0);

  octree_ = OctreeNode::Ptr(new OctreeNode(center, block_size_, nullptr, 0));
}

void VoxelMapOctree::initialize() {
  for (auto &&info : voxels_) {
    auto block = info.second;
  }
}

void VoxelMapOctree::querySegments(PointCloudXYZ::Ptr feats, Frame *frame,
                                   vector<Segment *> &segments) {
  segments.clear();

  const SE3 &Twb = frame->getTwb();

  // PointCloudXYZI::Ptr cloud(new PointCloudXYZI);
  // pcl::transformPointCloud(*feats, *cloud, Twb.matrix().cast<float>());

  // for (auto &&pt : cloud->points)
  std::vector<OctreeNode *> nodes_to_check;
  unordered_map<Cell *, Segment *> record;

  // std::vector<Segment *> tmp_segment;
  for (size_t i = 0; i < feats->size(); i++) {
    // auto &&Tpt = cloud->points[i];
    PointXYZ pt = feats->points[i];

    // Vector3d Tptx(Tpt.x, Tpt.y, Tpt.z);
    Vector3d ptx(pt.x, pt.y, pt.z);
    Vector3d Tptx = Twb * ptx;

    // TODO: leverage info from last frame for fast query
    VoxelIndex voxel_idx =
        getGridIndexFromPoint<VoxelIndex>(Tptx, block_size_inv_);
    auto hash_idx = AnyIndexHash()(voxel_idx);

    if (!voxels_.count(hash_idx)) {
      // voxels_[hash_idx] = Block::Ptr(
      //     new Block(block_size_, voxel_idx.cast<double>() * block_size_));
      continue;
    }
    Block::Ptr block = voxels_[hash_idx];

    // oct
    auto node = block->octree_->getLeafForPoint(Tptx);

    if (!node->cell_) {
      continue;
    }

    Segment *segment;
    if (!record.count(node->cell_)) {
      segment = new Segment();
      segment->cell_ = node->cell_;

      // tmp_segment.push_back(segment);

      segment->frame_ = frame;
      record[node->cell_] = segment;
      // node->cell_->addObservation(frame, segment);

      // node->cell_->
    } else {
      segment = record[node->cell_];
    }

    segment->cachePoint(Tptx, ptx);

    // if (!node->cell_->optable_) {
    //   // STEP. cache cell to check list
    //   nodes_to_check.push_back(node);
    //   // nodes_to_check
    // }
  }

  for (auto &&info : record) {
    auto &&segment = info.second;
    // if (segment->pts_.size() < mapping::ba::min_num_seg_surf ||
    //     segment->gaussian_.scalar.x() >= 0.1) {
    //   segment->cell_->frames_.pop_back();
    //   segment->cell_->segments_.pop_back();
    //   delete segment;
    //   continue;
    // }

    segments.push_back(segment);
    segment->estimateParam();
    segment->clearCache();
    // if (segment->gaussian_.scalar.x() < 0.1) {
    //   segments.push_back(segment);
    // } else {
    //   segment->cell_->frames_.pop_back();
    //   segment->cell_->segments_.pop_back();
    //   delete segment;
    // }
  }

  // INFOLOG("current size for voxels: {}", voxels_.size());
}

void VoxelMapOctree::insertFrame(PointCloudXYZ::Ptr feats, Frame *frame,
                                 vector<Segment *> &segments, bool b_add_obs) {
  segments.clear();

  const SE3 &Twb = frame->getTwb();

  // PointCloudXYZI::Ptr cloud(new PointCloudXYZI);
  // pcl::transformPointCloud(*feats, *cloud, Twb.matrix().cast<float>());

  // for (auto &&pt : cloud->points)
  std::vector<OctreeNode *> nodes_to_check;
  unordered_map<Cell *, Segment *> record;

  // std::vector<Segment *> tmp_segment;
  for (size_t i = 0; i < feats->size(); i++) {
    // auto &&Tpt = cloud->points[i];
    PointXYZ pt = feats->points[i];

    // Vector3d Tptx(Tpt.x, Tpt.y, Tpt.z);
    Vector3d ptx(pt.x, pt.y, pt.z);
    Vector3d Tptx = Twb * ptx;

    // TODO: leverage info from last frame for fast query
    VoxelIndex voxel_idx =
        getGridIndexFromPoint<VoxelIndex>(Tptx, block_size_inv_);
    auto hash_idx = AnyIndexHash()(voxel_idx);

    if (!voxels_.count(hash_idx)) {
      voxels_[hash_idx] = Block::Ptr(
          new Block(block_size_, voxel_idx.cast<double>() * block_size_));
    }
    Block::Ptr block = voxels_[hash_idx];

    // oct
    auto node = block->octree_->getLeafForPoint(Tptx);

    if (!node->cell_) {
      node->cell_ = new Cell;
    }

    Segment *segment;
    if (!record.count(node->cell_)) {
      segment = new Segment();
      segment->cell_ = node->cell_;

      // tmp_segment.push_back(segment);

      segment->frame_ = frame;
      if (b_add_obs) {
        record[node->cell_] = segment;
        node->cell_->addObservation(frame, segment);
      }
      // node->cell_->
    } else {
      segment = record[node->cell_];
    }

    segment->cachePoint(Tptx, ptx);

    // if (!node->cell_->optable_) {
    //   // STEP. cache cell to check list
    //   nodes_to_check.push_back(node);
    //   // nodes_to_check
    // }
  }

  for (auto &&info : record) {
    auto &&segment = info.second;
    // if (segment->pts_.size() < mapping::ba::min_num_seg_surf ||
    //     segment->gaussian_.scalar.x() >= 0.1) {
    //   segment->cell_->frames_.pop_back();
    //   segment->cell_->segments_.pop_back();
    //   delete segment;
    //   continue;
    // }

    // if (ba::b_map_check_num && segment->pts_.size() < ba::th_map_min_num) {
    //   segment->cell_->frames_.pop_back();
    //   segment->cell_->segments_.pop_back();
    //   delete segment;
    //   continue;
    // }
    bool c4 = ba::b_map_check_num && segment->pts_.size() < ba::th_map_min_num;
    if (c4) {
      segment->cell_->frames_.pop_back();
      segment->cell_->segments_.pop_back();
      delete segment;

      continue;
    }

    // segments.push_back(segment);
    segment->estimateParam();
    // TODO: how to separate estimation?
    // segment->estimateParamHomo();

    segment->clearCache();

    Vector3d scalar = segment->gaussian_.scalar;
    // cout << scalar.x() << endl;
    // INFOLOG("num: {}, scalar: {}, th: {}", segment->num_, scalar.x(),
    //         test_ba::th_map_min_lambda * test_ba::voxel_size_surf);
    // bool c
    bool c1 = ba::b_map_check_min_lambda &&
              (segment->gaussian_.scalar.x() > ba::th_map_min_lambda);

    double max_lambda =
        sqrt(segment->gaussian_.scalar.y() + segment->gaussian_.scalar.z());
    bool c2 = ba::b_map_check_max_lambda &&
              (max_lambda < common::resolution * ba::th_map_max_lambda);

    double p2 = (scalar.y() - scalar.x()) / scalar.z();
    bool c3 = ba::b_map_check_p2 && (p2 < ba::th_map_p2);

    // segment->gaussian_.scalar.x() < test_ba::th_map_max_lambda;
    if (c1 || c2 || c3 || c4) {
      segment->cell_->frames_.pop_back();
      segment->cell_->segments_.pop_back();
      delete segment;
    } else {
      segments.push_back(segment);
    }

    // if (segment->gaussian_.scalar.x() < 0.1) {
    //   segments.push_back(segment);
    // } else {
    //   segment->cell_->frames_.pop_back();
    //   segment->cell_->segments_.pop_back();
    //   delete segment;
    // }
  }

  // INFOLOG("current size for voxels: {}", voxels_.size());
}

void VoxelMapOctree::insertFrameEF(PointCloudXYZ::Ptr feats, Frame *frame,
                                   vector<Segment *> &segments,
                                   bool b_add_obs) {
  segments.clear();

  const SE3 &Twb = frame->getTwb();

  std::vector<OctreeNode *> nodes_to_check;
  unordered_map<Cell *, Segment *> record;

  for (size_t i = 0; i < feats->size(); i++) {
    PointXYZ pt = feats->points[i];

    Vector3d ptx(pt.x, pt.y, pt.z);
    Vector3d Tptx = Twb * ptx;

    // TODO: leverage info from last frame for fast query
    VoxelIndex voxel_idx =
        getGridIndexFromPoint<VoxelIndex>(Tptx, block_size_inv_);
    auto hash_idx = AnyIndexHash()(voxel_idx);

    if (!voxels_.count(hash_idx)) {
      voxels_[hash_idx] = Block::Ptr(
          new Block(block_size_, voxel_idx.cast<double>() * block_size_));
    }
    Block::Ptr block = voxels_[hash_idx];

    // oct
    auto node = block->octree_->getLeafForPoint(Tptx);

    if (!node->cell_) {
      node->cell_ = new Cell;
    }

    Segment *segment;
    if (!record.count(node->cell_)) {
      segment = new Segment();
      segment->cell_ = node->cell_;

      segment->frame_ = frame;
      if (b_add_obs) {
        record[node->cell_] = segment;
        node->cell_->addObservation(frame, segment);
      }
    } else {
      segment = record[node->cell_];
    }

    segment->cachePoint(Tptx, ptx);
  }

  for (auto &&info : record) {
    auto &&segment = info.second;
    if (segment->pts_.size() < 10 || segment->gaussian_.scalar.x() >= 0.1) {
      segment->cell_->frames_.pop_back();
      segment->cell_->segments_.pop_back();
      delete segment;
      continue;
    }

    segments.push_back(segment);
    segment->estimateParam();
    // TODO: how to separate estimation?
    segment->estimateParamHomo();

    segment->clearCache();

    // if (segment->gaussian_.scalar.x() < 0.1) {
    //   segments.push_back(segment);
    // } else {
    //   segment->cell_->frames_.pop_back();
    //   segment->cell_->segments_.pop_back();
    //   delete segment;
    // }
  }

  // INFOLOG("current size for voxels: {}", voxels_.size());
}

void VoxelMapOctree::clearInactiveVoxels(const Vector3d &pt_max,
                                         const Vector3d &pt_min) {
  vector<size_t> indices;
  for (auto &&info : voxels_) {
    auto &&idx = info.first;
    auto &&voxel_ptr = info.second;

    Vector3d centre =
        voxel_ptr->origin_ + voxel_ptr->block_size_ * 0.5 * Vector3d::Ones();

    if (centre.x() > pt_max.x() || centre.y() > pt_max.y() ||
        centre.z() > pt_max.z() || centre.x() < pt_min.x() ||
        centre.y() < pt_min.y() || centre.z() < pt_min.z()) {
      indices.push_back(idx);
    }
  }
  INFOLOG("#to_clear: {}", indices.size());
  for (auto &&idx : indices) {
    voxels_.erase(idx);
  }

  // throw std::runtime_error("not implemented");
}

OctreeNode *VoxelMapOctree::queryPoints(const Vector3d &pt3d) {
  // STEP.0 get hash idx
  VoxelIndex idx = getGridIndexFromPoint<VoxelIndex>(pt3d, block_size_inv_);

  auto hash_idx = AnyIndexHash()(idx);

  if (!voxels_.count(hash_idx)) {
    return nullptr;
  } else {
    auto block = voxels_[hash_idx];
    CHECK_NOTNULL(block->octree_);

    return block->octree_->getNodeForPoint(pt3d);
  }
}

}  // namespace xreg
