#include "xreg/visualization/voxel_map_visualizer.h"

namespace xreg {

using namespace std;

VoxelMapVisualizer::VoxelMapVisualizer(ros::NodeHandle &nh,
                                       VoxelMapOctree::Ptr map) {
  map_ = map;
  topic_ = "voxel_map";

  pub_ = nh.advertise<visualization_msgs::MarkerArray>(topic_, 10, true);
  cloud_pub_ = nh.advertise<sensor_msgs::PointCloud2>("voxel_cloud", 10, true);

  // msg_ = ;
}

// void VoxelMapVisualizer::setQueryResult(
//     const std::vector<OctreeNode *> &nodes) {
//   query_results_ = unordered_set<OctreeNode *>(nodes.begin(), nodes.end());
// }

void VoxelMapVisualizer::visualize() {
  static int seq = 0;
  CHECK_NOTNULL(map_);

  if (!msg_) {
    msg_ = visualization_msgs::MarkerArray::Ptr(
        new visualization_msgs::MarkerArray);
  }

  msg_->markers.clear();

  auto stamp = ros::Time::now();
  auto frame_id = "map";

  int count = 0;
  // auto marker = visualization_msgs::Marker();
  // marker.header.frame_id = frame_id;
  // marker.header.seq = 0;
  // marker.header.stamp = stamp;

  // construct voxel msgs
  auto insertVoxel = [&](const Vector3d &origin, const double size,
                         const Vector4d &color, const string &ns) {
    // Vector3d origin = v.second->origin_;
    // double size = v.second->block_size_;
    // INFOLOG("origin: {}", origin.transpose());
    auto marker = visualization_msgs::Marker();
    marker.header.frame_id = frame_id;
    marker.header.seq = 0;
    marker.header.stamp = stamp;

    marker.id = count++;
    marker.ns = ns;
    marker.type = visualization_msgs::Marker::LINE_LIST;
    marker.action = visualization_msgs::Marker::ADD;
    marker.scale.x = 0.04;
    marker.color.a = 0.8;
    marker.color.r = color(1);
    marker.color.g = color(2);
    marker.color.b = color(3);

    marker.pose.position.x = 0.0;
    marker.pose.position.y = 0.0;
    marker.pose.position.z = 0.0;
    marker.pose.orientation.x = 0.0;
    marker.pose.orientation.y = 0.0;
    marker.pose.orientation.z = 0.0;
    marker.pose.orientation.w = 1.0;

    geometry_msgs::Point point0, point1, point2;
    std_msgs::ColorRGBA color_line;
    color_line.a = 0.8;
    color_line.r = color(1);
    color_line.g = color(2);
    color_line.b = color(3);

    vector<geometry_msgs::Point> points(8);
    points[0].x = origin.x();
    points[0].y = origin.y();
    points[0].z = origin.z();
    points[1].x = origin.x() + size;
    points[1].y = origin.y();
    points[1].z = origin.z();
    points[2].x = origin.x();
    points[2].y = origin.y() + size;
    points[2].z = origin.z();
    points[3].x = origin.x();
    points[3].y = origin.y();
    points[3].z = origin.z() + size;
    points[4].x = origin.x() + size;
    points[4].y = origin.y() + size;
    points[4].z = origin.z();
    points[5].x = origin.x() + size;
    points[5].y = origin.y();
    points[5].z = origin.z() + size;
    points[6].x = origin.x();
    points[6].y = origin.y() + size;
    points[6].z = origin.z() + size;
    points[7].x = origin.x() + size;
    points[7].y = origin.y() + size;
    points[7].z = origin.z() + size;
    marker.points.push_back(points[0]);
    marker.points.push_back(points[1]);
    marker.points.push_back(points[0]);
    marker.points.push_back(points[2]);
    marker.points.push_back(points[4]);
    marker.points.push_back(points[1]);
    marker.points.push_back(points[4]);
    marker.points.push_back(points[2]);

    marker.points.push_back(points[3]);
    marker.points.push_back(points[5]);
    marker.points.push_back(points[3]);
    marker.points.push_back(points[6]);
    marker.points.push_back(points[7]);
    marker.points.push_back(points[5]);
    marker.points.push_back(points[7]);
    marker.points.push_back(points[6]);

    marker.points.push_back(points[0]);
    marker.points.push_back(points[3]);
    marker.points.push_back(points[1]);
    marker.points.push_back(points[5]);
    marker.points.push_back(points[2]);
    marker.points.push_back(points[6]);
    marker.points.push_back(points[4]);
    marker.points.push_back(points[7]);

    msg_->markers.push_back(marker);
  };

  Vector4d block_color(0.8, 0.1, 0.1, 0.8);
  Vector4d leave_color(0.8, 0.8, 0.1, 0.1);

  sensor_msgs::PointCloud2::Ptr msg_cloud(new sensor_msgs::PointCloud2);
  pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud(
      new pcl::PointCloud<pcl::PointXYZRGB>);

  for (auto &&block : map_->voxels_) {
    auto &&idx = block.first;
    auto &&block_ptr = block.second;

    if (!block_ptr->octree_)
      continue;

    insertVoxel(block_ptr->origin_, block_ptr->block_size_, block_color,
                "block");

    deque<OctreeNode *> nodes = {block_ptr->octree_.get()};

    while (!nodes.empty()) {
      auto node = nodes.front();

      // Vector3d origin = node->
      // :w
      insertVoxel(node->getOrigin(), node->getSize(), node->color_,
                  to_string(node->depth_));

      if (query_results_.count(node)) {
        insertVoxel(node->getOrigin(), node->getSize(),
                    Vector4d(1.0, 1.0, 0.0, 0.0), "query");
      }

      if (node->is_leaf_) {
        // cout << node->getOrigin().transpose() << endl;
        // cout << node->getSize() << endl;
        insertVoxel(node->getOrigin(), node->getSize(), node->color_, "leaves");

        // for (auto &&ptx : node->points_) {
        //   pcl::PointXYZRGB pt;
        //   pt.x = ptx.x();
        //   pt.y = ptx.y();
        //   pt.z = ptx.z();
        //   pt.r = node->color_.y() * 255;
        //   pt.g = node->color_.z() * 255;
        //   pt.b = node->color_.w() * 255;

        //   cloud->points.push_back(pt);
        // }
      }

      nodes.pop_front();

      for (int i = 0; i < 8; i++) {
        auto child = node->getChild(i);
        if (child) {
          nodes.push_back(child);
        }
      }
    }

    // auto block_ptr
  }

  //   pcl::toROSMsg(*cloud, *msg_cloud);
  //   msg_cloud->header.frame_id = "map";
  //   msg_cloud->header.stamp = ros::Time::now();

  pub_.publish(msg_);
  //   cloud_pub_.publish(msg_cloud);
}
} // namespace xreg
