#include "xreg/optimization/factors.h"

#include "xreg/common/common.h"

using namespace xreg;
using namespace std;

int main(int argc, char **argv) {
  /* code */
  //   Eigen::Ve
  srand((unsigned int)time(0));
  Vector4d qi_coef = Vector4d::Random().normalized();
  //   cout << qi_coef << endl;
  Quaterniond qi(qi_coef);
  INFOLOG("qi: {}", qi_coef.transpose());
  INFOLOG("qi: {}", qi.coeffs());

  Vector4d qj_coef = Vector4d::Random().normalized();
  Quaterniond qj(qj_coef);

  Vector4d qji_coef = Vector4d::Random().normalized();
  Quaterniond qji(qji_coef);
  //   Vector4d qj = Vector4d::Random().normalized();

  Vector3d ti = Vector3d::Random();
  Vector3d tj = Vector3d::Random();
  Vector3d tji = Vector3d::Random();

  SE3 Tji(qji, tji);
  SE3 Ti(qi, ti);
  SE3 Tj(qj, tj);
  SO3 Tji_so3 = Tji.so3();
  // test rot_i
  Vector3d roti_log = Ti.so3().log();
  Vector3d rotj_log = Tj.so3().log();
  Vector3d roti_data = roti_log;
  Vector3d rotj_data = rotj_log;

  {
    ceres::CostFunction *factor = new RotationFactor(Tji_so3);
    double *params[2];
    double *jacob_arr[2];
    Vector3d residual;
    Eigen::Matrix<double, 3, 3, Eigen::RowMajor> jacobi, jacobj, jacobi_num,
        jacobj_num;

    params[0] = roti_data.data();
    params[1] = rotj_data.data();

    jacob_arr[0] = jacobi.data();
    jacob_arr[1] = jacobj.data();

    // Eigen::Matrix3d
    constexpr double eps = 1e-6;
    // for (size_t i = 0; i < 3; i++) {
    factor->Evaluate(params, residual.data(), jacob_arr);

    cout << residual.transpose() << endl;

    // STEP. rot 2 roti
    for (size_t i = 0; i < 3; i++) {
      Vector3d res_tmp;
      Vector3d delta_rot = Vector3d::Zero();
      delta_rot[i] += eps;
      roti_data = (Ti.so3() * SO3::exp(delta_rot)).log();

      factor->Evaluate(params, res_tmp.data(), nullptr);
      jacobi_num.col(i) = (res_tmp - residual) / eps;
      cout << res_tmp.transpose() << endl;
    }
    // INFOLOG("mat:\n{}\n\n{}", jacobi, jacobi_num);
    roti_data = Ti.so3().log();
    CHECK_LE((jacobi - jacobi_num).sum(), eps);

    // STEP. rot 2 rotj
    for (size_t i = 0; i < 3; i++) {
      Vector3d res_tmp;
      Vector3d delta_rot = Vector3d::Zero();
      delta_rot[i] += eps;
      rotj_data = (Tj.so3() * SO3::exp(delta_rot)).log();

      factor->Evaluate(params, res_tmp.data(), nullptr);
      jacobj_num.col(i) = (res_tmp - residual) / eps;
      cout << res_tmp.transpose() << endl;
    }
    rotj_data = Tj.so3().log();
    INFOLOG("mat:\n{}\n\n{}", jacobj, jacobj_num);
    Eigen::Matrix3d mat = (jacobj - jacobj_num);
    cout << mat << endl;
    CHECK_LE(mat.sum(), eps);

    delete factor;
  }

  {
    ceres::CostFunction *factor = new TranslationFactor(Tji);
    double *params[3];
    double *jacob_arr[3];
    Vector3d residual;
    Eigen::Matrix<double, 3, 3, Eigen::RowMajor> jacob_roti, jacob_ti, jacob_tj,
        jacob_roti_num, jacob_ti_num, jacob_tj_num;

    Vector3d ti_data = ti, tj_data = tj;
    roti_data = Ti.so3().log();
    params[0] = roti_data.data();
    params[1] = ti_data.data();
    params[2] = tj_data.data();

    jacob_arr[0] = jacob_roti.data();
    jacob_arr[1] = jacob_ti.data();
    jacob_arr[2] = jacob_tj.data();

    // Eigen::Matrix3d
    constexpr double eps = 1e-6;
    // for (size_t i = 0; i < 3; i++) {
    factor->Evaluate(params, residual.data(), jacob_arr);

    cout << residual.transpose() << endl;

    // STEP. rot 2 roti
    for (size_t i = 0; i < 3; i++) {
      Vector3d res_tmp;
      Vector3d delta_rot = Vector3d::Zero();
      delta_rot[i] += eps;
      roti_data = (Ti.so3() * SO3::exp(delta_rot)).log();

      factor->Evaluate(params, res_tmp.data(), nullptr);
      jacob_roti_num.col(i) = (res_tmp - residual) / eps;
      cout << res_tmp.transpose() << endl;
    }
    roti_data = Ti.so3().log();
    INFOLOG("mat:\n{}\n\n{}", jacob_roti, jacob_roti_num);
    CHECK_LE((jacob_roti - jacob_roti_num).sum(), eps);

    // STEP. rot 2 ti
    for (size_t i = 0; i < 3; i++) {
      Vector3d res_tmp;
      Vector3d delta_rot = Vector3d::Zero();
      delta_rot[i] += eps;
      ti_data = ti + delta_rot;

      factor->Evaluate(params, res_tmp.data(), nullptr);
      jacob_ti_num.col(i) = (res_tmp - residual) / eps;
      cout << res_tmp.transpose() << endl;
    }
    ti_data = ti;
    CHECK_LE((jacob_ti - jacob_ti_num).sum(), eps);

    // STEP. rot 2 ti
    for (size_t i = 0; i < 3; i++) {
      Vector3d res_tmp;
      Vector3d delta_rot = Vector3d::Zero();
      delta_rot[i] += eps;
      tj_data = tj + delta_rot;

      factor->Evaluate(params, res_tmp.data(), nullptr);
      jacob_tj_num.col(i) = (res_tmp - residual) / eps;
      cout << res_tmp.transpose() << endl;
    }
    tj_data = tj;
    CHECK_LE((jacob_tj - jacob_tj_num).sum(), eps);
    // INFOLOG("mat:\n{}\n\n{}", jacobi, jacobi_num);
    // CHECK_LE((jacobj - jacobj_num).sum(), eps);

    delete factor;
  }

  return 0;
}